from dask.diagnostics import Profiler, ResourceProfiler, CacheProfiler
from dask.diagnostics import visualize
import dask.threaded
from distributed import LocalCluster
from xicam.core import msg
from appdirs import user_config_dir
from .daskexecutor import DaskExecutor


class LocalExecutor(DaskExecutor):
    def execute(self, wf, client=None):
        if not client:
            client = dask.threaded
        return super(LocalExecutor, self).execute(wf, client)



import collections
import json
from urllib import parse
import os, sys
import pkg_resources
import requests
import yaml
from appdirs import user_config_dir, site_config_dir, user_cache_dir
import platform
import subprocess
from xicam.core import msg
from importlib import reload
import pip
from pathlib import Path

from . import venvs


def pipmain(args):
    old_env = os.environ.copy()
    os.environ["PYTHONPATH"] = os.path.join(venvs.current_environment, "Lib/site-packages")
    os.environ["PYTHONBASE"] = venvs.current_environment
    old_prefix = sys.prefix
    sys.prefix = venvs.current_environment

    r = 1  # (error code)
    try:
        # Install the bundled software

        try:
            r = pip.main(args)
        except AttributeError:
            from pip._internal import main

            r = main(args)
    except Exception as ex:
        msg.logError(ex)

    finally:
        os.environ = old_env
        sys.prefix = old_prefix
        return r

    # subprocess.call([os.path.join(venvs.current_environment, 'Scripts/pip.exe'), *args], env=env)
    # return 0  # Above subprocess typicall returns error code even though it succeeds


op_sys = platform.system()
if op_sys == "Darwin":  # User config dir incompatible with venv on darwin (space in path name conflicts)
    user_package_registry = Path(user_cache_dir(appname="xicam")) / "packages.yml"
else:
    user_package_registry = Path(user_config_dir(appname="xicam")) / "packages.yml"
site_package_registry = Path(site_config_dir(appname="xicam")) / "packages.yml"


def install(name: str):
    """
    Install a Xi-cam plugin package by querying the Xi-cam package repository with REST.

    Packages are installed into the currently active virtualenv

    Parameters
    ----------
    name : str
        The package name to be installed.
    """
    # TODO: test if installed
    # TODO: check if package is in repo

    # Get install plugin package information from cam-mart repository
    o = requests.get(f'http://cam.lbl.gov:5000/pluginpackages?where={{"name":"{name}"}}')

    # Get the uri from the plugin package information
    uri = parse.urlparse(json.loads(o.content)["_items"][0]["installuri"])

    failure = True

    print("Installing to:", venvs.current_environment)

    # Install from the uri
    if uri.scheme == "pipgit":  # Clones a git repo and installs with pip
        failure = pipmain(
            [
                "install",
                # f'--target={os.path.join(venvs.current_environment,"Lib/site-packages")}',
                "git+https://" + "".join(uri[1:]),
            ]
        )
    elif uri.scheme == "pip":
        failure = pipmain(
            [
                "install",
                # f'--target={os.path.join(venvs.current_environment,"Lib/site-packages")}',
                "".join(uri[1:]),
            ]
        )
    elif uri.scheme == "conda":
        raise NotImplementedError

    if not failure:
        pkg_registry[name] = uri.scheme

    from xicam.plugins import manager

    manager.collectPlugins()


def uninstall(name: str):
    # Note: pkg_resources keeps a "working_set" that won't normally be updated when venv changes...
    # An update may need to be forcefully triggered in the future
    pkg_resources._initialize_master_working_set()
    failure = True
    if name in pkg_registry:
        scheme = pkg_registry[name]
        if scheme in ["pipgit", "pip"]:
            failure = pipmain(["uninstall", "-y", name])
        elif scheme == "conda":
            raise NotImplementedError
    else:
        # TODO: Handle if name is not in the registry
        pass

    if not failure:
        del pkg_registry[name]

    return not failure


class pkg_registry(collections.MutableMapping):
    def __init__(self):
        self._store = dict()
        self.update(self._store)  # use the free update to set keys
        self.load()
        self.save()

    def __getitem__(self, key):
        return self._store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self._store[self.__keytransform__(key)] = value
        self.save()

    def __delitem__(self, key):
        del self._store[self.__keytransform__(key)]
        self.save()

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __keytransform__(self, key):
        return key

    def load(self):
        try:
            with open(user_package_registry, "r") as f:
                self._store = yaml.load(f.read())
        except FileNotFoundError:
            pass

    def save(self):
        user_package_registry.parent.mkdir(parents=True, exist_ok=True)
        with open(user_package_registry, "w") as f:
            f.write(yaml.dump(self._store))


pkg_registry = pkg_registry()

# def check_registry():
#
# def scan_for_unregistered():
#     unregistered_packages = {}
#     pkgs = pip.get_installed_distributions()
#     for pkg in pkgs:
#         targets = [pathlib.Path(pkg._provider.egg_info) / 'installed-files.txt',
#                    pathlib.Path(pkg._provider.egg_info) / 'RECORD']
#         for target in targets:
#             if target.exists():
#                 with open(str(target),'r') as f:
#                     if 'yapsy-plugin' in f.read():
#                         unregistered_packages[str(pathlib.Path(pkg.location)/pkg.key)] = pkg.project_name
#
#     return unregistered_packages



from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QGridLayout, QSplitter, QToolBar, QWidget


# TODO: this could be more generic, defining a collapsible interface/mixin type class
class CollapsibleWidget(QWidget):
    """
    Creates a widget that can be collapsed when a button is clicked.
    """

    toggled = Signal(bool)

    def __init__(self, widget: QWidget, buttonText: str, parent=None):
        """
        Constructs a widget that lets the passed ``widget`` keep an internal collapsed state that can be triggered when
        a button is clicked.

        Internally, when the button is clicked, a toggled signal is emitted, indicating what the collapse state has
        been toggled to. Additionally, this signal is connected to the collapse() slot, which will collapse the passed
        widget if another widget has been added via addWidget(). The widget added via addWidget() is not collapsible.

        Parameters
        ----------
        widget
            The widget to make collapsible.
        buttonText
            The text of the button that will be used to collapse.
        parent
            The parent widget.
        """
        super(CollapsibleWidget, self).__init__(parent)
        self.widget = widget
        self.buttonText = buttonText
        self.collapsed = False

        toolBar = QToolBar()
        action = toolBar.addAction(self.buttonText, self.toggle)
        action.setIconText("&" + action.text())
        self.collapseButton = toolBar.widgetForAction(action)
        self.collapseButton.setCheckable(True)
        self.collapseButton.setChecked(not self.collapsed)

        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.addWidget(self.widget)
        self.splitter.setCollapsible(0, self.collapsed)
        # Keep track of the collapsed widget's size to restore properly when un-collapsed
        self.widgetSize = self.splitter.sizes()[0]

        layout = QGridLayout()
        layout.addWidget(self.splitter, 0, 0)
        layout.addWidget(toolBar, 1, 0)

        self.setLayout(layout)

        self.toggled.connect(self.collapse)

    def addWidget(self, widget):
        """
        Adds a non-collapsible widget to the internal splitter.

        Parameters
        ----------
        widget
            Non-collapsible widget to add.
        """
        # TODO -- what happens when more than one widget is added?
        self.splitter.addWidget(widget)
        self.splitter.setCollapsible(1, False)

    def toggle(self):
        self.collapsed = not self.collapsed
        self.toggled.emit(self.collapsed)

    def collapse(self, collapsed):
        self.collapseButton.setChecked(not collapsed)
        self.splitter.setCollapsible(0, collapsed)
        # Only do something for now if there is more than one widget added.
        if len(self.splitter.sizes()) > 1:
            if collapsed:
                self.widgetSize = self.splitter.sizes()[0]
                sizes = [0, self.splitter.sizes()[1]]
                self.splitter.setSizes(sizes)
            else:
                sizes = [self.widgetSize, self.splitter.sizes()[1] - self.widgetSize]
                self.splitter.setSizes(sizes)



"""
Experimental Qt-based data browser for bluesky
"""
import ast
from datetime import datetime
import event_model
import functools
import itertools
import jsonschema
import logging
import queue
import threading
import time

from qtpy.QtCore import Qt, Signal, QThread, QSettings
from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtWidgets import (
    QAbstractItemView,
    QPushButton,
    QComboBox,
    QDateTimeEdit,
    QHeaderView,
    QHBoxLayout,
    QMessageBox,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
    QTableView,
    QMenu,
    QApplication
)

from xicam.core import msg

from .utils import ConfigurableQObject
from .top_utils import load_config, Callable
from xicam.core import msg
from xicam.core import threads

MAX_SEARCH_RESULTS = 100  # TODO Use fetchMore instead of a hard limit.
log = logging.getLogger("bluesky_browser")
BAD_TEXT_INPUT = """
QLineEdit {
    background-color: rgb(255, 100, 100);
}
"""
GOOD_TEXT_INPUT = """
QLineEdit {
    background-color: rgb(255, 255, 255);
}
"""
RELOAD_INTERVAL = 11
_validate = functools.partial(jsonschema.validate, types={"array": (list, tuple)})


def timeit(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        print("{:s} function took {:.3f} ms".format(f.__name__, (time2 - time1) * 1000.0))
        return ret

    return wrap


def default_search_result_row(entry):
    start = entry.metadata["start"]
    stop = entry.metadata["stop"]
    start_time = datetime.fromtimestamp(start["time"])
    if stop is None:
        str_duration = "-"
    else:
        duration = datetime.fromtimestamp(stop["time"]) - start_time
        str_duration = str(duration)
        str_duration = str_duration[: str_duration.index(".")]
    return {
        "Unique ID": start["uid"][:8],
        "Transient Scan ID": (start.get("scan_id", "-")),
        "Plan Name": start.get("plan_name", "-"),
        "Start Time": start_time.strftime("%Y-%m-%d %H:%M:%S"),
        "Duration": str_duration,
        "Exit Status": "-" if stop is None else stop["exit_status"],
    }

# TODO: implement threads with event-loops

class ReloadThread(QThread):
    def __init__(self, search_state):
        super(ReloadThread, self).__init__()
        self.search_state = search_state
        QApplication.instance().aboutToQuit.connect(self.quit)

    def run(self):
        while True:
            t0 = time.monotonic()
            # Never reload until the last reload finished being
            # displayed.
            self.search_state.show_results_event.wait()
            # Wait for RELOAD_INTERVAL to pass or until we are poked,
            # whichever happens first.
            self.search_state.reload_event.wait(max(0, RELOAD_INTERVAL - (time.monotonic() - t0)))
            self.search_state.reload_event.clear()
            # Reload the catalog to show any new results.
            self.search_state.reload()


class ProcessQueriesThread(QThread):
    def __init__(self, search_state):
        super(ProcessQueriesThread, self).__init__()
        self.search_state = search_state
        QApplication.instance().aboutToQuit.connect(self.quit)

    def run(self):
        while True:
            try:
                self.search_state.process_queries()
            except Exception as e:
                msg.logError(e)
                msg.showMessage("Unable to query: ", str(e))


class SearchState(ConfigurableQObject):
    """
    Encapsulates CatalogSelectionModel and SearchResultsModel. Executes search.
    """

    new_results_catalog = Signal([])
    sig_update_header = Signal()
    search_result_row = Callable(default_search_result_row, config=True)

    def __init__(self, catalog):
        self.last_results_thread = None
        self.update_config(load_config())
        self.selected_catalog = None
        self.root_catalog = self.flatten_remote_catalogs(catalog)
        self.enabled = False  # to block searches during initial configuration
        self.catalog_selection_model = CatalogSelectionModel()
        self.search_results_model = SearchResultsModel(self)
        self._subcatalogs = []  # to support lookup by item's positional index
        self.open_uids = set()  # to support quick lookup that a catalog is open
        self._results_catalog = None
        self._new_uids_queue = queue.Queue(maxsize=MAX_SEARCH_RESULTS)  # to pass reference of new catalogs across threads
        self.list_subcatalogs()
        self.set_selected_catalog(0)
        self.query_queue = queue.Queue()
        self.show_results_event = threading.Event()
        self.show_results_event.set()
        self.reload_event = threading.Event()

        super().__init__()

        self.new_results_catalog.connect(self.start_show_results)

        self.reload_thread = ReloadThread(self)
        self.reload_thread.start()
        self.process_queries_thread = ProcessQueriesThread(self)
        self.process_queries_thread.start()

    def start_show_results(self):
        self.last_results_thread = threads.QThreadFuture(self.show_results)
        self.last_results_thread.start()

    def flatten_remote_catalogs(self, catalog):
        from intake.catalog.base import Catalog

        cat_dict = {}

        for name in catalog:
            try:
                from intake.catalog.base import RemoteCatalog

                sub_cat = catalog[name]
                # @TODO remote catalogs are one level too high. This check is
                # pretty rough. Would rather check that a catalog's children
                # should be top-level.
                # This is making the terrible assumption that children
                # of a RemoteCatalog be treated as top-level. But until
                # we figure out how to tell that a catalog is a real catalog
                # with data, it's as good as we can get
                if isinstance(sub_cat(), RemoteCatalog):
                    for name in sub_cat:
                        cat_dict[name] = sub_cat[name]
                else:
                    cat_dict[name] = sub_cat
            except Exception as e:
                msg.logError(e)
                msg.showMessage("Unable to query top level catalogs: ", str(e))

        return Catalog.from_dict(cat_dict)

    def request_reload(self):
        self._results_catalog.force_reload()
        self.reload_event.set()

    def apply_search_result_row(self, entry):
        try:
            return self.search_result_row(entry)
        except Exception as exc:
            # Either the documents in entry are not valid or the definition of
            # search_result_row (which will be user-configurable) has failed to
            # account for some possiblity. Figure out which situation this is.
            try:
                _validate(entry.metadata["start"], event_model.schemas[event_model.DocumentNames.start])
            except jsonschema.ValidationError:
                log.exception("Invalid RunStart Document: %r", entry.metadata["start"])
                raise SkipRow("invalid document") from exc
            try:
                _validate(entry.metadata["stop"], event_model.schemas[event_model.DocumentNames.stop])
            except jsonschema.ValidationError:
                if entry.metadata["stop"] is None:
                    log.debug("Run %r has no RunStop document.", entry.metadata["start"]["uid"])
                else:
                    log.exception("Invalid RunStop Document: %r", entry.metadata["stop"])
                raise SkipRow("invalid document")
            log.exception("Run with uid %s raised error with search_result_row.", entry.metadata["start"]["uid"])
            raise SkipRow("error in search_result_row") from exc

    def __del__(self):
        if hasattr(self.reload_thread):
            self.reload_thread.terminate()

    def list_subcatalogs(self):
        self._subcatalogs.clear()
        self.catalog_selection_model.clear()
        if not self.root_catalog:
            return

        for name in self.root_catalog:
            self._subcatalogs.append(name)
            self.catalog_selection_model.appendRow(QStandardItem(str(name)))

    def set_selected_catalog(self, item):
        if len(self._subcatalogs) == 0:
            return
        name = self._subcatalogs[item]
        try:
            self.selected_catalog = self.root_catalog[name]()
            self.search()
        except Exception as e:
            log.error(e)
            msg.showMessage("Unable to contact catalog: ", str(e))

    def check_for_new_entries(self):
        # check for any new results and add them to the queue for later processing
        found_new = False
        for uid in itertools.islice(self._results_catalog, MAX_SEARCH_RESULTS):
            if uid in self.open_uids:
                continue
            self.open_uids.add(uid)
            self._new_uids_queue.put(uid)
            found_new = True
        return found_new

    def process_queries(self):
        # If there is a backlog, process only the newer query.
        block = True

        while True:
            try:
                query = self.query_queue.get_nowait()
                block = False
            except queue.Empty:
                if block:
                    query = self.query_queue.get()
                break
        log.debug("Submitting query %r", query)
        try:
            t0 = time.monotonic()
            msg.showMessage("Running Query")
            msg.showBusy()
            if not self.selected_catalog:
                return
            self._results_catalog = self.selected_catalog.search(query)
            found_new = self.check_for_new_entries()
            duration = time.monotonic() - t0
            log.debug("Query yielded %r results (%.3f s).", len(self._results_catalog), duration)
            if found_new and self.show_results_event.is_set():
                self.new_results_catalog.emit()
        except Exception as e:
            msg.logError(e)
            msg.showMessage("Problem running query")
        finally:
            msg.hideBusy()

    def search(self):
        self._new_uids_queue = queue.Queue(maxsize=MAX_SEARCH_RESULTS)
        if self.last_results_thread and not self.show_results_event.is_set():
            self.last_results_thread.requestInterruption()
            self.last_results_thread.wait()
        self.search_results_model.clear()
        self.search_results_model.selected_rows.clear()
        self.open_uids.clear()
        if not self.enabled:
            return
        query = {"time": {}}
        if self.search_results_model.since is not None:
            query["time"]["$gte"] = self.search_results_model.since
        if self.search_results_model.until is not None:
            query["time"]["$lt"] = self.search_results_model.until
        query.update(**self.search_results_model.custom_query)
        self.query_queue.put(query)

    @timeit
    def get_run_by_uid(self, uid):
        try:
            return self._results_catalog[uid]
        except Exception:
            raise SkipRow(f"error accessing documents for {uid}")

    def show_results(self):
        header_labels_set = False
        self.show_results_event.clear()
        t0 = time.monotonic()
        counter = 0

        try:
            msg.showBusy()
            while not self._new_uids_queue.empty():
                counter += 1
                row = []
                new_uid = self._new_uids_queue.get()
                try:
                    entry = self.get_run_by_uid(new_uid)
                    row_data = self.apply_search_result_row(entry)
                except SkipRow as e:
                    msg.showMessage(str(msg))
                    msg.logError(e)
                    continue
                if not header_labels_set:
                    # Set header labels just once.
                    threads.invoke_in_main_thread(self.search_results_model.setHorizontalHeaderLabels, list(row_data))
                    header_labels_set = True
                for value in row_data.values():
                    item = QStandardItem()
                    item.setData(value, Qt.DisplayRole)
                    item.setData(new_uid, Qt.UserRole)
                    row.append(item)
                if QThread.currentThread().isInterruptionRequested():
                    self.show_results_event.set()
                    msg.logMessage("Interrupt requested")
                    return
                threads.invoke_in_main_thread(self.search_results_model.appendRow, row)
            if counter:
                self.sig_update_header.emit()
                duration = time.monotonic() - t0
                msg.showMessage("Displayed {} new results {}.".format(counter, duration))
            self.show_results_event.set()
        except Exception as e:
            msg.showMessage("Error displaying runs")
            msg.logError(e)
        finally:
            msg.hideBusy()

    def reload(self):
        t0 = time.monotonic()
        if self._results_catalog is not None:
            try:
                self._results_catalog.reload()
                new_results = self.check_for_new_entries()
                duration = time.monotonic() - t0
                msg.logMessage("Reloaded search results {}.".format(duration))
                if new_results and self.show_results_event.is_set():
                    self.new_results_catalog.emit()
            except Exception as e:
                log.error(e)
                msg.showMessage("Unable to query top level catalogs: ", str(e))


class CatalogSelectionModel(QStandardItemModel):
    """
    List the subcatalogs in the root Catalog.
    """

    ...


class SearchResultsModel(QStandardItemModel):
    """
    Perform searches on a Catalog and model the results.
    """

    selected_result = Signal([list])
    open_entries = Signal([str, list])
    preview_entry = Signal(str, object)
    valid_custom_query = Signal([bool])

    def __init__(self, search_state, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.custom_query = {}
        self.search_state = search_state
        self.since = None
        self.until = None
        self.selected_rows = set()

    def emit_selected_result(self, selected, deselected):
        try:
            self.selected_rows |= set(index.row() for index in selected.indexes())
            self.selected_rows -= set(index.row() for index in deselected.indexes())
            entries = []
            for row in sorted(self.selected_rows):
                uid = self.data(self.index(row, 0), Qt.UserRole)
                entry = self.search_state._results_catalog[uid]
                entries.append(entry)
            self.selected_result.emit(entries)
        except Exception as e:
            msg.logError(e)
            msg.showMessage("Problem getting info about for selected row")

    def emit_open_entries(self, target, indexes):
        rows = set(index.row() for index in indexes)
        entries = []
        for row in rows:
            uid = self.data(self.index(row, 0), Qt.UserRole)
            entry = self.search_state._results_catalog[uid]
            entries.append(entry)
        self.open_entries.emit(target, entries)

    def emit_preview_entry(self, target, index):
        row = index.row()
        entries = []
        uid = self.data(self.index(row, 0), Qt.UserRole)
        entry = self.search_state._results_catalog[uid]
        entries.append(entry)
        self.preview_entry.emit(target, entry)

    def on_search_text_changed(self, text):
        try:
            self.custom_query = dict(ast.literal_eval(text)) if text else {}
        except Exception:
            self.valid_custom_query.emit(False)
        else:
            self.valid_custom_query.emit(True)
            self.search_state.search()

    def on_since_time_changed(self, datetime):
        self.since = datetime.toSecsSinceEpoch()
        self.search_state.search()

    def on_until_time_changed(self, datetime):
        self.until = datetime.toSecsSinceEpoch()
        self.search_state.search()


class SearchInputWidget(QWidget):
    """
    Input fields for specifying searches on SearchResultsModel
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_bar = QLineEdit()
        search_bar_layout = QHBoxLayout()
        search_bar_layout.addWidget(QLabel("Custom Query:"))
        search_bar_layout.addWidget(self.search_bar)
        mongo_query_help_button = QPushButton()
        mongo_query_help_button.setText("?")
        search_bar_layout.addWidget(mongo_query_help_button)
        mongo_query_help_button.clicked.connect(self.show_mongo_query_help)

        self.since_widget = QDateTimeEdit()
        self.since_widget.setCalendarPopup(True)
        self.since_widget.setDisplayFormat("yyyy-MM-dd HH:mm")
        since_layout = QHBoxLayout()
        since_layout.addWidget(QLabel("Since:"))
        since_layout.addWidget(self.since_widget)

        self.until_widget = QDateTimeEdit()
        self.until_widget.setCalendarPopup(True)
        self.until_widget.setDisplayFormat("yyyy-MM-dd HH:mm")
        until_layout = QHBoxLayout()
        until_layout.addWidget(QLabel("Until:"))
        until_layout.addWidget(self.until_widget)

        layout = QVBoxLayout()
        layout.addLayout(since_layout)
        layout.addLayout(until_layout)
        layout.addLayout(search_bar_layout)
        self.setLayout(layout)

    def mark_custom_query(self, valid):
        "Indicate whether the current text is a parsable query."
        if valid:
            stylesheet = GOOD_TEXT_INPUT
        else:
            stylesheet = BAD_TEXT_INPUT
        self.search_bar.setStyleSheet(stylesheet)

    def show_mongo_query_help(self):
        "Launch a Message Box with instructions for custom queries."
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText("For advanced search capability, enter a valid Mongo query.")
        msg.setInformativeText(
            """
Examples:

{'plan_name': 'scan'}
{'proposal': 1234},
{'$and': ['proposal': 1234, 'sample_name': 'Ni']}
"""
        )
        msg.setWindowTitle("Custom Mongo Query")
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()


class CatalogList(QComboBox):
    """
    List of subcatalogs
    """

    ...


class CatalogSelectionWidget(QWidget):
    """
    Input widget for selecting a subcatalog
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.catalog_list = CatalogList()
        layout = QHBoxLayout()
        layout.addWidget(QLabel("Catalog:"))
        layout.addWidget(self.catalog_list)
        self.setLayout(layout)


class SearchResultsWidget(QTableView):
    """
    Table of search results
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.setSortingEnabled(True)
        self.setSelectionBehavior(QTableView.SelectRows)
        self.setShowGrid(False)
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignHCenter)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.setAlternatingRowColors(True)

    def hide_hidden_columns(self):
        hidden_columns = QSettings().value("catalog.columns.hidden") or set()
        header = self.horizontalHeader()
        current_column_names = [str(self.model().headerData(i, Qt.Horizontal)) for i in range(header.count())]
        current_hidden_names = hidden_columns.intersection(set(current_column_names))
        for name in current_hidden_names:
            header.setSectionHidden(current_column_names.index(name), True)


class SearchWidget(QWidget):
    """
    Search input and results list
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.catalog_selection_widget = CatalogSelectionWidget()
        self.search_input_widget = SearchInputWidget()
        self.search_results_widget = SearchResultsWidget()
        layout = QVBoxLayout()
        layout.addWidget(self.catalog_selection_widget)
        layout.addWidget(self.search_input_widget)
        layout.addWidget(self.search_results_widget)
        self.setLayout(layout)

        header = self.search_results_widget.horizontalHeader()
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(self.header_menu)

    def _current_column_names(self, header):
        return [self._current_column_name(header, i) for i in range(header.count())]

    def _current_column_name(self, header, index):
        return str(header.model().headerData(index, Qt.Horizontal))

    def hide_column(self, header, logicalIndex):
        """
        Hide a column, adding the column from the list of hidden columns in QSettings
        """
        hidden_columns = QSettings().value("catalog.columns.hidden") or set()
        if len(hidden_columns) == header.count() - 1:
            msg.notifyMessage("Only one column is left to hide, cannot hide all of them.")
            return
        hidden_columns.add(self._current_column_name(header, logicalIndex))
        QSettings().setValue("catalog.columns.hidden", hidden_columns)
        header.setSectionHidden(logicalIndex, True)

    def unhide_column(self, header, logicalIndex):
        """
        Unhide a column, removing the column from the list of hidden columns in QSettings
        """
        hidden_columns = QSettings().value("catalog.columns.hidden") or set()
        column_name = self._current_column_name(header, logicalIndex)
        try:
            hidden_columns.remove(column_name)
        except KeyError as ex:
            raise (KeyError(f"Attempted to unhide non-hidden column name {column_name}."))
        QSettings().setValue("catalog.columns.hidden", hidden_columns)
        header.setSectionHidden(logicalIndex, False)

    def header_menu(self, position):
        """
        Creates a menu allowing users to show and hide columns
        """
        header = self.sender()  # type: QHeaderView
        index = header.logicalIndexAt(position)
        menu = QMenu("Options")
        action = menu.addAction("Hide Column")  # type: QAction
        column_name = str(header.model().headerData(index, Qt.Horizontal))
        action.triggered.connect(lambda: self.hide_column(header, index))
        show_columns_menu = menu.addMenu("Show Columns")

        for i in range(header.count()):
            if header.isSectionHidden(i):
                column_name = str(header.model().headerData(i, Qt.Horizontal))
                action = show_columns_menu.addAction(column_name)
                action.triggered.connect(functools.partial(self.unhide_column, header, i))
                # why does below work, but not: lambda: self.unhide_column(header, i)
                # action.triggered.connect(lambda triggered, logicalIndex=i: self.unhide_column(header, logicalIndex))

        menu.exec_(header.mapToGlobal(position))


class SkipRow(Exception):
    ...



from .plugin import PluginType
import uuid
import datetime
from typing import Tuple, List
from xicam.core.data import lazyfield
from pathlib import Path

# Note: Split into DataHandlerPlugin and IngestorPlugin?


class DataHandlerPlugin(PluginType):
    """
    This base class defines a reader/writer for an on-disk file format. This interface will be structured such that the
    format definition is registered with FabIO at activation, and will mirror the FabIO API structure. Subclass
    instances should not depend on other plugins. Example: A reader/writer for the *.fits file format.

    See the FabIO API for a detailed explanation of the file format abstraction.

    See xicam.plugins.tests for examples.

    """

    is_singleton = False
    needs_qt = False

    DESCRIPTION = ""

    DEFAULT_EXTENTIONS = []

    MAGIC_NUMBERS = []

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def getStartDoc(cls, paths, start_uid):
        metadata = cls.parseTXTFile(paths[0])
        metadata.update(cls.parseDataFile(paths[0]))
        descriptor_keys = getattr(cls, "descriptor_keys", [])
        metadata = dict([(key, metadata.get(key)) for key in descriptor_keys])
        return start_doc(start_uid=start_uid)

    @classmethod
    def getEventDocs(cls, paths, descriptor_uid):
        shape = cls(paths[0])().shape  # Assumes each frame has same shape
        for path in paths:
            metadata = cls.parseTXTFile(path)
            metadata.update(cls.parseDataFile(path))
            yield embedded_local_event_doc(descriptor_uid, "primary", cls, (path,), metadata=metadata)

    @staticmethod
    def getDescriptorUIDs(paths):
        return str(uuid.uuid4())

    @classmethod
    def getDescriptorDocs(cls, paths, start_uid, descriptor_uid):
        metadata = cls.parseTXTFile(paths[0])
        metadata.update(cls.parseDataFile(paths[0]))

        metadata = dict([(key, metadata.get(key, None)) for key in getattr(cls, "descriptor_keys", [])])
        yield descriptor_doc(start_uid, descriptor_uid, metadata=metadata)

    @classmethod
    def getStopDoc(cls, paths, start_uid):
        return stop_doc(start_uid=start_uid)

    @classmethod
    def reduce_paths(cls, paths):
        return paths

    @classmethod
    def title(cls, paths):
        if len(paths) > 1:
            return f"Series: {Path(paths[0]).resolve().stem}…"
        return Path(paths[0]).resolve().stem

    @classmethod
    def _setTitle(cls, startdoc, paths):
        startdoc["sample_name"] = cls.title(paths)
        return startdoc

    @classmethod
    def ingest(cls, paths):
        paths = cls.reduce_paths(paths)
        start_uid = str(uuid.uuid4())
        descriptor_uids = cls.getDescriptorUIDs(paths)
        return {
            "start": cls._setTitle(cls.getStartDoc(paths, start_uid), paths),
            "descriptors": list(cls.getDescriptorDocs(paths, start_uid, descriptor_uids)),
            "events": list(cls.getEventDocs(paths, descriptor_uids)),
            "stop": cls.getStopDoc(paths, start_uid),
        }

    def parseTXTFile(self, *args, **kwargs):
        return {}

    def parseDataFile(self, *args, **kwargs):
        return {}


def start_doc(start_uid: str, metadata: dict = None):
    if not metadata:
        metadata = {}
    metadata.update({"uid": start_uid, "time": datetime.datetime.now().strftime("%a %b %d %H:%M:%S %Y")})
    return metadata


def event_doc(data_uid: str, descriptor_uid: str, metadata: dict = None):
    if not metadata:
        metadata = {}
    metadata.update(
        {
            "descriptor": descriptor_uid,
            "time": datetime.datetime.now().strftime("%a %b %d %H:%M:%S %Y"),
            "uid": str(uuid.uuid4()),
            "data": {"primary": data_uid},
        }
    )
    return metadata


def embedded_local_event_doc(
    descriptor_uid: str,
    field: str,
    handler: type,
    resource_path: tuple = None,
    resource_kwargs: dict = None,
    metadata: dict = None,
):
    if not resource_kwargs:
        resource_kwargs = {}
    if not metadata:
        metadata = {}

    datafield = {field: lazyfield(handler, resource_path, resource_kwargs)}
    metadata.update(
        FillableDict(
            {"descriptor": descriptor_uid, "time": datetime.datetime.now(), "uid": str(uuid.uuid4()), "data": datafield}
        )
    )
    return metadata


def descriptor_doc(start_uid: str, descriptor_uid: str, metadata: dict = None):
    if not metadata:
        metadata = {}
    metadata.update({"run_start": start_uid, "name": "primary", "uid": descriptor_uid})
    return metadata


def stop_doc(start_uid: str, metadata: dict = None):
    if not metadata:
        metadata = {}
    metadata.update(
        {
            "run_start": start_uid,
            "time": 0,  # TODO: set this to the cumulative time of the full doc
            "uid": str(uuid.uuid4()),
            "exit_status": "success",
        }
    )
    return metadata


class FillableDict(dict):
    def __init__(self, *args, **kwargs):
        super(FillableDict, self).__init__(*args, **kwargs)
        self.filled = False

    def fill(self):
        self.update({"data": self["data"]["handler"](*self["data"]["args"], **self["data"]["kwargs"])})
        self.filled = True



from xicam.plugins.datahandlerplugin import DataHandlerPlugin, start_doc
import numpy as np


class NPYPlugin(DataHandlerPlugin):
    name = "NPYPlugin"

    DEFAULT_EXTENTIONS = [".npy"]

    def __call__(self, *args, **kwargs):
        return np.load(self.path, allow_pickle=True)

    def __init__(self, path):
        super(NPYPlugin, self).__init__()
        self.path = path

    @classmethod
    def getStartDoc(cls, paths, start_uid):
        return start_doc(start_uid=start_uid, metadata={"paths": paths})



from pathlib import Path
from typing import Union, List, Set

from qtpy.QtCore import Signal, QObject

from databroker.utils import ALL
from databroker.in_memory import BlueskyInMemoryCatalog
from warnings import warn

import mimetypes
import warnings

import entrypoints


def detect_mimetypes(filename: str) -> List[str]:
    """
    Take in a filename; return a mimetype string like 'image/tiff'.
    """
    # First rely on custom "sniffers" that can employ file signatures (magic
    # numbers) or any other format-specific tricks to extract a mimetype.
    matched_mimetypes = list()

    with open(filename, "rb") as file:
        # The choice of 64 bytes is arbitrary. We may increase this in the
        # future if we discover reason to. Therefore, sniffers should not
        # assume that they will receive this exact number of bytes.
        first_bytes = file.read(64)
    for ep in entrypoints.get_group_all("databroker.sniffers"):
        try:
            content_sniffer = ep.load()  # TODO: only load sniffers once
        except Exception as ex:
            msg.logError(ex)
        else:
            matched_mimetype = content_sniffer(filename, first_bytes)
            if matched_mimetype:
                matched_mimetypes.append(matched_mimetype)

    # Guessing the mimetype from the mimemtype db is quick, lets do it always
    matched_mimetype = mimetypes.guess_type(filename)[0]
    if matched_mimetype:
        matched_mimetypes.append(matched_mimetype)

    if not matched_mimetypes:
        raise UnknownFileType(f"Could not identify the MIME type of {filename}")

    return matched_mimetypes


def applicable_ingestors(filename, mimetype):
    """
    Take in a filename and its mimetype; return a list of compatible ingestors.
    """
    # Find ingestor(s) for this mimetype.
    ingestors = []
    for ep in entrypoints.get_group_all("databroker.ingestors"):
        if ep.name == mimetype:
            try:
                ingestor = ep.load()
            except Exception as ex:
                msg.logError(ex)
            else:
                ingestors.append(ingestor)

    return ingestors


def choose_ingestor(filename, mimetype):
    """
    Take in a filename and its mimetype; return an ingestor.
    If multiple are found, return the first one (deterministic but arbitrary).
    If none are found, raise.
    """
    ingestors = applicable_ingestors(filename, mimetype)
    if not ingestors:
        raise NoIngestor(f"No ingestors were applicable to {filename}")
    elif len(ingestors) > 1:
        warnings.warn("More than one ingestor was applicable.")
    return ingestors[0]


class UnknownFileType(ValueError):
    ...


class NoIngestor(ValueError):
    ...


def load_header(uris: List[Union[str, Path]] = None, uuid: str = None):
    """
    Load a document object, either from a file source or a databroker source, by uuid. If loading from a filename, the
    file will be registered in databroker.

    Parameters
    ----------
    uris
    uuid

    Returns
    -------
    NonDBHeader

    """
    from xicam.plugins import manager as pluginmanager  # must be a late import

    # ext = Path(filename).suffix[1:]
    # for cls, extensions in extension_map.items():
    #     if ext in extensions:

    # First try to see if we have a databroker ingestor, then fall-back to Xi-cam DataHandlers
    ingestor = None
    filename = str(Path(uris[0]))

    # Sanity checks
    if Path(filename).is_dir():
        msg.logMessage("Opening dir; nothing to load.")
        return

    if not Path(filename).exists():
        raise FileExistsError(f"Attempted to load non-existent file: {filename}")

    try:
        mimetypes = detect_mimetypes(filename)
    except UnknownFileType as ex:
        msg.logError(ex)
        mimetypes = []
    else:
        msg.logMessage(f"Mimetypes detected: {mimetypes}")

        # TODO: here, we try each valid mimetype; some GUI for selection will be needed

        for mimetype in mimetypes:
            try:
                ingestor = choose_ingestor(filename, mimetype)
            except NoIngestor as e:
                pass
            else:
                msg.logMessage(f"Ingestor selected: {ingestor}")
                break

    if ingestor:
        document = list(ingestor(uris))
        uid = document[0][1]["uid"]
        catalog = BlueskyInMemoryCatalog()
        # TODO -- change upsert signature to put start and stop as kwargs
        # TODO -- ask about more convenient way to get a BlueskyRun from a document generator
        catalog.upsert(document[0][1], document[-1][1], ingestor, [uris], {})
        return catalog[uid]
    else:
        warn(f"No applicable ingestor found. Falling-back to DataHandlers")

    handlercandidates = []
    ext = Path(uris[0]).suffix
    for plugin in pluginmanager.get_plugins_of_type("DataHandlerPlugin"):
        if ext in plugin.DEFAULT_EXTENTIONS:
            handlercandidates.append(plugin)
    if not handlercandidates:
        return NonDBHeader({}, [], [], {})
    # try:
    msg.logMessage(f"Handler selected: {handlercandidates[0]}")
    return NonDBHeader(**handlercandidates[0].ingest(uris))
    # except (IsADirectoryError, TypeError):
    #     # TODO: add Header ingestor for directory
    #     return NonDBHeader({}, [], [], {})


class NonDBHeader(object):
    """
    A dictionary-like object summarizing metadata for a run.
    """

    # def _repr_html_(self):
    #     env = jinja2.Environment()
    #     env.filters['human_time'] = _pretty_print_time
    #     template = env.from_string(_HTML_TEMPLATE)
    #     return template.render(document=self)

    ### dict-like methods ###

    def __init__(self, start: dict = None, descriptors: List[dict] = None, events: List[dict] = None, stop: dict = None):
        self._documents = {
            "start": [start] if start else [],
            "descriptor": descriptors or [],
            "event": events or [],
            "stop": [stop] if stop else [],
        }

    def append(self, docname, doc):
        if docname in ["start", "stop"] and self._documents[docname]:
            raise KeyError(f"A {docname} document already exists within this header.")
        self._documents[docname].append(doc)

    def __getitem__(self, k):
        try:
            return getattr(self, k)
        except AttributeError as e:
            raise KeyError(k)

    @property
    def startdoc(self):
        startdocs = self._documents["start"]
        return startdocs[0] if startdocs else {}

    @property
    def eventdocs(self):
        return self._documents["event"]

    @property
    def descriptordocs(self):
        return self._documents["descriptor"]

    def get(self, *args, **kwargs):
        return getattr(self, *args, **kwargs)

    def items(self):
        for k in self.keys():
            yield k, getattr(self, k)

    def values(self):
        for k in self.keys():
            yield getattr(self, k)

    def keys(self):
        yield from iter(self._documents.keys())

    # def to_name_dict_pair(self):
    #     ret = attr.asdict(self)
    #     ret.pop('db')
    #     ret.pop('_cache')
    #     ret['descriptors'] = self.descriptors
    #     return self._name, ret

    def __len__(self):
        return 4

    def __iter__(self):
        return self.keys()

    # ## convenience methods and properties, encapsulating one-liners ## #

    @property
    def descriptors(self):
        yield from self._documents["descriptor"]

    @property
    def stream_names(self):
        raise NotImplementedError

    def fields(self, stream_name=ALL):
        """
        Return the names of the fields ('data keys') in this run.

        Parameters
        ----------
        stream_name : string or ``ALL``, optional
            Filter results by stream name (e.g., 'primary', 'baseline'). The
            default, ``ALL``, combines results from all streams.

        Returns
        -------
        fields : set

        Examples
        --------
        Load the most recent run and list its fields.

        >>> h = db[-1]
        >>> h.fields()
        {'eiger_stats1_total', 'eiger_image'}

        See Also
        --------
        :meth:`Header.devices`
        """
        return {key for event in self["eventdocs"] for key in event["data"].keys()}

    def devices(self, stream_name=ALL):
        """
        Return the names of the devices in this run.

        Parameters
        ----------
        stream_name : string or ``ALL``, optional
            Filter results by stream name (e.g., 'primary', 'baseline'). The
            default, ``ALL``, combines results from all streams.

        Returns
        -------
        devices : set

        Examples
        --------
        Load the most recent run and list its devices.

        >>> h = db[-1]
        >>> h.devices()
        {'eiger'}

        See Also
        --------
        :meth:`Header.fields`
        """
        result = set()
        for d in self.descriptors:
            if stream_name is ALL or stream_name == d.get("name", "primary"):
                result.update(d["object_keys"])
        return result

    def config_data(self, device_name):
        """
        Extract device configuration data from Event Descriptors.

        This refers to the data obtained from ``device.read_configuration()``.

        See example below. The result is structed as a [...deep breath...]
        dictionary of lists of dictionaries because:

        * The device might have been read in multiple event streams
          ('primary', 'baseline', etc.). Each stream name is a key in the
          outer dictionary.
        * The configuration is typically read once per event stream, but in
          general may be read multiple times if the configuration is changed
          mid-stream. Thus, a list is needed.
        * Each device typically produces multiple configuration fields
          ('exposure_time', 'period', etc.). These are the keys of the inner
          dictionary.

        Parameters
        ----------
        device_name : string
            device name (originally obtained from the ``name`` attribute of
            some readable Device)

        Returns
        -------
        result : dict
            mapping each stream name (such as 'primary' or 'baseline') to a
            list of data dictionaries

        Examples
        --------
        Get the device configuration recorded for the device named 'eiger'.

        >>> h.config_data('eiger')
        {'primary': [{'exposure_time': 1.0}]}

        Assign the exposure time to a variable.

        >>> exp_time = h.config_data('eiger')['primary'][0]['exposure_time']

        How did we know that ``'eiger'`` was a valid argument? We can query for
        the complete list of device names:

        >>> h.device_names()
        {'eiger', 'cs700'}
        """
        result = defaultdict(list)
        for d in sorted(self.descriptors, key=lambda d: d["time"]):
            config = d["configuration"].get(device_name)
            if config:
                result[d["name"]].append(config["data"])
        return dict(result)  # strip off defaultdict behavior

    def documents(self, stream_name=ALL, fields=None, fill=False):
        """
        Load all documents from the run.

        This is a generator the yields ``(name, doc)``.

        Parameters
        ----------
        stream_name : string or ``ALL``, optional
            Filter results by stream name (e.g., 'primary', 'baseline'). The
            default, ``ALL``, combines results from all streams.
        fill : bool, optional
            Whether externally-stored data should be filled in. False by
            default.

        Yields
        ------
        name, doc : (string, dict)

        Examples
        --------
        Loop through the documents from a run.

        >>> h = db[-1]
        >>> for name, doc in h.headers():
        ...     # do something
        """
        for docname in self.keys():
            for doc in self._documents[docname]:
                yield docname, doc

    def stream(self, *args, **kwargs):
        warn("The 'stream' method been renamed to 'documents'. The old name " "will be removed in the future.")
        yield from self.documents(*args, **kwargs)

    def table(self, stream_name="primary", fields=None, fill=False, timezone=None, convert_times=True, localize_times=True):
        """
        Load the data from one event stream as a table (``pandas.DataFrame``).

        Parameters
        ----------
        stream_name : str, optional
            Get events from only "event stream" with this name.

            Default is 'primary'

        fields : List[str], optional
            whitelist of field names of interest; if None, all are returned

            Default is None

        fill : bool or Iterable[str], optional
            Which fields to fill.  If `True`, fill all
            possible fields.

            Each event will have the data filled for the intersection
            of it's external keys and the fields requested filled.

            Default is False

        handler_registry : dict, optional
            mapping filestore specs (strings) to handlers (callable classes)

        convert_times : bool, optional
            Whether to convert times from float (seconds since 1970) to
            numpy datetime64, using pandas. True by default.

        timezone : str, optional
            e.g., 'US/Eastern'; if None, use metadatastore configuration in
            `self.mds.config['timezone']`

        localize_times : bool, optional
            If the times should be localized to the 'local' time zone.  If
            True (the default) the time stamps are converted to the localtime
            zone (as configure in mds).

            This is problematic for several reasons:

              - apparent gaps or duplicate times around DST transitions
              - incompatibility with every other time stamp (which is in UTC)

            however, this makes the dataframe repr look nicer

            This implies convert_times.

            Defaults to True to preserve back-compatibility.

        Returns
        -------
        table : pandas.DataFrame

        Examples
        --------
        Load the 'primary' data stream from the most recent run into a table.

        >>> h = db[-1]
        >>> h.table()

        This is equivalent. (The default stream_name is 'primary'.)

        >>> h.table(stream_name='primary')
                                    time intensity
        0  2017-07-16 12:12:37.239582345       102
        1  2017-07-16 12:12:39.958385283       103

        Load the 'baseline' data stream.

        >>> h.table(stream_name='baseline')
                                    time temperature
        0  2017-07-16 12:12:35.128515999         273
        1  2017-07-16 12:12:40.128515999         274
        """
        raise NotImplementedError

    def events(self, stream_name="primary", fields=None, fill=False):
        """
        Load all Event documents from one event stream.

        This is a generator the yields Event documents.

        Parameters
        ----------
        stream_name : str, optional
            Get events from only "event stream" with this name.

            Default is 'primary'

        fields : List[str], optional
            whitelist of field names of interest; if None, all are returned

            Default is None

        fill : bool or Iterable[str], optional
            Which fields to fill.  If `True`, fill all
            possible fields.

            Each event will have the data filled for the intersection
            of it's external keys and the fields requested filled.

            Default is False

        Yields
        ------
        doc : dict

        Examples
        --------
        Loop through the Event documents from a run. This is 'lazy', meaning
        that only one Event at a time is loaded into memory.

        >>> h = db[-1]
        >>> for event in h.events():
        ...    # do something

        List the Events documents from a run, loading them all into memory at
        once.

        >>> events = list(h.events())
        # """
        # ev_gen = self.db.get_events([self], stream_name=stream_name,
        #                             fields=fields, fill=fill)

        for ev in self._documents["event"]:
            if not set(fields).isdisjoint(set(ev["data"].keys())) or not fields:
                yield ev

    def data(self, field, stream_name="primary", fill=True):
        """
        Extract data for one field. This is convenient for loading image data.

        Parameters
        ----------
        field : string
            such as 'image' or 'intensity'

        stream_name : string, optional
            Get data from a single "event stream." Default is 'primary'

        fill : bool, optional
             If the data should be filled.

        Yields
        ------
        data
        """
        if fill:
            fill = {field}
        for event in self.events(stream_name=stream_name, fields=[field], fill=fill):
            yield event["data"][field].asarray()

    def meta_array(self, field=None):
        return DocMetaArray(self, field)


class QNonDBHeader(QObject, NonDBHeader):
    sigChanged = Signal()

    def append(self, docname, doc):
        super(QNonDBHeader, self).append(docname, doc)
        self.sigChanged.emit()


from functools import lru_cache
import numpy as np
from xicam.core import msg


class MetaXArray(object):
    def __init__(self, dataarray):
        self.dataarray = dataarray

        for attr in ["size", "ndim", "shape", "dtype"]:
            setattr(self, attr, getattr(self.dataarray, attr))

    def __len__(self):
        return self.shape[0]

    def xvals(self):
        return range(len(self))

    def min(self):
        return self.dtype.type(self.dataarray.min())

    def max(self):
        return self.dtype.type(self.dataarray.max())

    def ravel(self):
        return self.dataarray.values.ravel()

    def __getitem__(self, item: Union[List[slice], int]):
        if isinstance(item, List):
            return np.asarray(self.dataarray[tuple(item)])
        else:
            return np.asarray(self.dataarray[item])

    def transpose(self, ax):
        if ax != [0, 1, 2]:
            raise ValueError(
                "A MetaXArray cannot actually be transposed; the transpose method is provided for "
                "compatibility with pyqtgraph"
                "s ImageView"
            )
        return self

    def view(self, _):
        return self

    def implements(self, _):
        return False


class DocMetaArray(object):
    def __init__(self, header: NonDBHeader, field: str = None):
        self._dtype = None
        self.header = header
        self._field = field
        self._events = None
        self._len = None
        self._shape = None
        self._size = None
        self._ndim = None

    @property
    def events(self):
        if not self._events:
            self._events = list(self.header.events(fields=[self.field]))
        return self._events

    def __len__(self):
        if not self._len:
            self._len = len(list(self.events))
        return self._len

    @property
    def size(self):
        if not self._size:
            self._size = self.slice(0).size * len(self)
        return self._size

    @property
    def ndim(self):
        if not self._ndim:
            self._ndim = self.slice(0).ndim + 1
        return self._ndim

    @property
    def shape(self):
        if not self._shape:
            self._shape = (len(self), *self.slice(0).shape)
        return self._shape

    @property
    def dtype(self):
        if not self._dtype:
            firstslice = self.slice(0)
            self._dtype = firstslice.dtype
        return self._dtype

    @property
    def field(self):
        if not self._field:
            fields = list(self.header.fields())
            if len(fields) > 1:
                msg.logError(
                    ValueError("Unspecified field for document stream with >1 field. Potentially unexpected behavior.")
                )
                self.field = next(iter(self.header.eventdocs[0]["data"].keys()))
            else:
                self.field = fields[0]

        return self._field

    @field.setter
    def field(self, value):
        self._field = value

    def min(self):
        if not self._min:
            # cache using slice
            self.slice(0)
        return self._min

    def max(self):
        if not self._max:
            # cache using slice
            self.slice(0)
        return self._max

    def slice(self, i):
        if not self.events:
            return None

        arr = self.events[i]["data"][self.field]
        if hasattr(arr, "asarray"):
            arr = arr.asarray()
        if not isinstance(arr, np.ndarray):
            arr = np.array(arr)
        self._min = arr.min()
        self._max = arr.max()
        return arr

    def __getitem__(self, item: Union[List[slice], int]):
        if isinstance(item, list) or isinstance(item, tuple) and len(item) > 1:
            rmin = item[0].start if item[0].start is not None else 0
            rmax = item[0].stop if item[0].stop is not None else self.shape[0]
            rstep = item[0].step if item[0].step is not None else 1
            r = range(rmin, rmax, rstep)
            return np.array([self.slice(i)[item[1:]] for i in r])
        return self.slice(item)

    def transpose(self, ax):
        # if ax != [0,1,2]:
        #     raise ValueError('A DocMetaArray cannot actually be transposed; the transpose method is provided for '
        #                      'compatibility with pyqtgraph''s ImageView')
        return self

    def view(self, _):
        return self


# TODO: Add asarray


class lazyfield(object):
    def __init__(self, handler_cls, resource_path, resource_kwargs):
        self.handler_cls = handler_cls
        self.resource_path = resource_path
        self.resource_kwargs = resource_kwargs
        self._handler = None
        super(lazyfield, self).__init__()

    @property
    def handler(self):
        if not self._handler:
            self._handler = self.handler_cls(*self.resource_path)
        return self._handler

    def implements(self, t):
        if t == "MetaArray":
            return True

    def asarray(self):
        return self.handler(**self.resource_kwargs)


# TODO: Eliminate lazyfield and use only handler in doc?



def test_msg():
    from xicam.core import msg

    msg.logMessage("this", "is", "a", "tests:", 42, level=msg.WARNING)


def test_exception_msg():
    from xicam.core import msg  # force load the message module
    import pytest

    # Raise an error; this should be
    with pytest.raises(Exception):
        raise RuntimeError("Something bad happened...")



from pyqtgraph.parametertree import Parameter, parameterTypes, ParameterTree
from qtpy.QtWidgets import QDialog, QVBoxLayout, QDialogButtonBox
from qtpy.QtCore import Qt

"""
Usage:

parameterized_scan = ParameterizablePlan(scan)

plan = parameterized_scan([det], device_list(), min, max)

plan_parameter = plan.parameter

RE(plan)

"""


class ParameterizablePlan(object):
    def __init__(self, plan):
        self.plan = plan

    def __call__(self, *args, **kwargs):
        return ParameterizedPlan(self.plan, args, kwargs)


class ParameterizedPlan(object):
    def __init__(self, plan, args, kwargs):
        self.plan = plan
        self.args = args
        self.kwargs = kwargs
        self._parameter = None

    @property
    def parameter(self):
        if not self._parameter:
            self._parameter = args_to_params(*self.args, **self.kwargs)
        return self._parameter

    def __iter__(self):
        args = list(self.args)
        kwargs = dict()
        for i, arg in enumerate(args):
            if isinstance(arg, Parameter):
                args[i] = arg.value()
            else:
                args[i] = arg

        for key, value in self.kwargs.items():
            if isinstance(value, Parameter):
                kwargs[key] = value.value()
            else:
                kwargs[key] = value

        return self.plan(*args, **kwargs)

    # def __next__(self):
    #     if not self._plan_instance:
    #         self._plan_instance = self.plan(*self.args, **self.kwargs)
    #     return next(self._plan_instance)


def parameterize(func):
    """
    Usage:

    from pyqtgraph.parametertree.parameterTypes import SimpleParameter

    @parameterize
    def sum(a, b):
        return sum(a, b)

    a = SimpleParameter(name='a', type='float', value=1)
    b = SimpleParameter(name='b', type='float', value=2)

    sum(a, b)  # spawns a dialog; user-interaction required
    """

    def func_wrapper(*args, **kwargs):
        param = args_to_params(*args, **kwargs)

        # query values
        paramdialog = ParameterDialog(param)
        paramdialog.exec_()

        # extract values
        args = list(args)
        for i, arg in enumerate(args):
            if isinstance(arg, Parameter):
                args[i] = arg.value()
            else:
                args[i] = arg

        for key, value in kwargs.items():
            if isinstance(value, Parameter):
                kwargs[key] = value.value()
            else:
                kwargs[key] = value

        # make the call
        return func(*args, **kwargs)

    return func_wrapper


def args_to_params(*args, **kwargs):
    param = parameterTypes.GroupParameter(name="Parameters")
    for arg in args + tuple(kwargs.values()):
        if isinstance(arg, Parameter):
            param.addChild(arg)
        elif isinstance(arg, (list, tuple)):
            param.addChild(args_to_params(*arg))
    return param


class ParameterDialog(QDialog):
    def __init__(self, children, parent=None):
        super(ParameterDialog, self).__init__(parent)

        layout = QVBoxLayout(self)

        self.paramtree = ParameterTree()
        layout.addWidget(self.paramtree)

        self.paramtree.setParameters(children, showTop=False)

        # OK and Cancel buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def exec_(self, *args, **kwargs):
        result = super(ParameterDialog, self).exec_()
        if result != self.Accepted:
            raise InterruptedError("Execution aborted by user.")



from warnings import warn

from qtpy.QtWidgets import QTabBar, QMenu, QAction, QTabWidget
from qtpy.QtGui import QStandardItemModel, QMouseEvent
from qtpy.QtCore import QItemSelectionModel, QObject, Qt
from typing import List
from functools import partial
from xicam.core import msg


class TabView(QTabWidget):
    def __init__(
        self,
        catalogmodel: QStandardItemModel = None,
        selectionmodel: QItemSelectionModel = None,
        widgetcls=None,
        stream=None,
        field=None,
        bindings: List[tuple] = [],
        **kwargs,
    ):
        """

        Parameters
        ----------
        model
        widgetcls
        stream
        field
        bindings
            A list of tuples with pairs of bindings, s.t. the one item is the name of the attribute on widget cls holding a
            signal (to be mirrored across each new widget), and the second is the receiver.
        kwargs
        """
        super(TabView, self).__init__()
        self.setTabBar(ContextMenuTabBar())
        self.kwargs = kwargs

        self.setWidgetClass(widgetcls)
        self.catalogmodel = None
        self.selectionmodel = None
        if selectionmodel:
            self.setSelectionModel(selectionmodel)  # type: TabItemSelectionModel

        if catalogmodel:
            self.setCatalogModel(catalogmodel)

        self.stream = stream
        self.field = field
        self.bindings = bindings

        self.setTabsClosable(True)
        self.setDocumentMode(True)
        self.tabCloseRequested.connect(self.closeTab)

    def setCatalogModel(self, model: QStandardItemModel):
        self.catalogmodel = model
        if not self.selectionmodel:
            self.selectionmodel = TabItemSelectionModel(self)
        model.dataChanged.connect(self.dataChanged)

    def setHeaderModel(self, model):
        self.setCatalogModel(model)
        warn(DeprecationWarning())

    def setSelectionModel(self, model: QItemSelectionModel):
        self.selectionmodel = model
        self.selectionmodel.currentChanged.connect(lambda current, previous: self.setCurrentIndex(current.row()))
        self.currentChanged.connect(
            lambda i: self.selectionmodel.setCurrentIndex(self.catalogmodel.index(i, 0), QItemSelectionModel.Rows)
        )
        self.currentChanged.connect(
            lambda _: self.selectionmodel.setCurrentIndex(
                self.catalogmodel.index(self.currentIndex(), 0), QItemSelectionModel.ClearAndSelect
            )
        )

    def dataChanged(self, start, end):
        for i in range(self.catalogmodel.rowCount()):
            itemdata = None
            if hasattr(self.catalogmodel.item(i), "header"):
                itemdata = self.catalogmodel.item(i).header
            else:
                itemdata = self.catalogmodel.item(i).data(Qt.UserRole)

            if self.widget(i):
                if (hasattr(self.widget(i), "catalog") and self.widget(i).catalog == itemdata) or (
                    hasattr(self.widget(i), "header") and self.widget(i).header == itemdata
                ):
                    continue
            try:
                newwidget = self.widgetcls(itemdata, stream=self.stream, field=self.field, **self.kwargs)
            except Exception as ex:
                msg.logMessage(
                    f"A widget of type {self.widgetcls} could not be initialized with args: {itemdata, self.field, self.kwargs}"
                )
                msg.logError(ex)
                self.catalogmodel.removeRow(i)
                self.dataChanged(0, 0)
                return

            self.setCurrentIndex(self.insertTab(i, newwidget, self.catalogmodel.item(i).text()))

            for sender, receiver in self.bindings:
                if isinstance(sender, str):
                    sender = getattr(newwidget, sender)
                if isinstance(receiver, str):
                    receiver = getattr(newwidget, receiver)
                sender.connect(receiver)

        for i in reversed(range(self.catalogmodel.rowCount(), self.count())):
            self.removeTab(i)

    def setWidgetClass(self, cls):
        self.widgetcls = cls

    def currentCatalog(self):
        return self.catalogmodel.item(self.currentIndex()).data(Qt.UserRole)

    def currentHeader(self):
        return self.catalogmodel.item(self.currentIndex()).header
        warn(DeprecationWarning())

    def closeTab(self, i):
        newindex = self.currentIndex()
        if i <= self.currentIndex():
            newindex -= 1

        self.removeTab(i)
        self.catalogmodel.removeRow(i)
        self.selectionmodel.setCurrentIndex(self.catalogmodel.index(newindex, 0), QItemSelectionModel.Rows)


class TabViewSynchronizer(QObject):
    def __init__(self, tabviews: List[TabView]):
        super(TabViewSynchronizer, self).__init__()
        self.tabviews = tabviews
        for tabview in tabviews:
            tabview.currentChanged.connect(partial(self.sync, sourcetabview=tabview))
            tabview.tabCloseRequested.connect(partial(self.sync, sourcetabview=tabview))

    def sync(self, index, sourcetabview):
        for tabview in self.tabviews:
            if tabview is sourcetabview:
                continue
            tabview.setCurrentIndex(index)
            tabview.dataChanged(None, None)


class ContextMenuTabBar(QTabBar):
    def __init__(self):
        super(ContextMenuTabBar, self).__init__()
        self.contextMenu = QMenu()
        self.closeaction = QAction("&Close")
        self.closeaction.triggered.connect(self.close)
        self.closeothersaction = QAction("Close &Others")
        self.closeothersaction.triggered.connect(self.closeothers)
        self.closeallaction = QAction("Close &All")
        self.closeallaction.triggered.connect(self.closeall)
        self.contextMenu.addActions([self.closeaction, self.closeothersaction, self.closeallaction])
        self._rightclickedtab = None

    def close(self):
        self.tabCloseRequested.emit(self._rightclickedtab)

    def closeothers(self):
        for i in reversed(range(self.count())):
            if i != self._rightclickedtab:
                self.tabCloseRequested.emit(i)

    def closeall(self):
        for i in reversed(range(self.count())):
            self.tabCloseRequested.emit(i)

    def mousePressEvent(self, event: QMouseEvent):
        super(ContextMenuTabBar, self).mousePressEvent(event)
        self._rightclickedtab = self.tabAt(event.pos())
        if self._rightclickedtab != -1:
            if event.button() == Qt.RightButton:
                self.contextMenu.popup(self.mapToGlobal(event.pos()))


class TabItemSelectionModel(QItemSelectionModel):
    def __init__(self, tabview: TabView):
        super(TabItemSelectionModel, self).__init__(tabview.catalogmodel)
        self.tabview = tabview

    def currentIndex(self):
        return self.tabview.currentIndex()



import json
from dask.diagnostics import Profiler, ResourceProfiler, CacheProfiler
from dask.diagnostics import visualize
import dask.threaded
from xicam.core import msg
from appdirs import user_config_dir

try:
    from camlink.services import graph as task_graph
except ImportError:
    print("CAMLINK is not installed; functionally coming soon™...")
import distributed
from .daskexecutor import DaskExecutor

client = None


class CamLinkExecutor(DaskExecutor):
    def execute(self, wf, client=None):
        # global client, graph

        services = {
            "machines": [
                {
                    "name": "freyja",
                    "address": "freyja.nsls2.bnl.gov",
                    "port": 22,
                    "username": "rp",
                    "password": "keyfile:/home/rp/.ssh/id_rsa",
                    "environment": {"PYTHONUSERBASE": "/tmp"},
                    "config_dir": "/tmp/camera",
                }
            ],
            "graph": {
                "configure": [{"machine": "freyja", "apps": [{"name": "dask/dask-scheduler"}], "tasks": ["dask-cluster"]}]
            },
        }

        services = json.dumps(services)
        print(services)

        graph = task_graph.Graph()
        graph.parse_stream(services)
        graph.start_tasks()
        graph.connect()
        graph.execute()

        meta_data = graph.machines[0].tasks[0].request_meta_data()

        local_port = graph.machines[0].node.get_free_local_port()
        remote_port = meta_data[0][0]
        print(local_port, remote_port)
        graph.machines[0].node.forward_tunnel(local_port, "localhost", remote_port)

        client = distributed.Client("tcp://localhost:" + str(local_port))

        return super(CamLinkExecutor, self).execute(wf, client)[0]



from typing import Dict, List
import copy
from itertools import count

import numpy as np
import pyqtgraph as pg
from qtpy.QtGui import QTransform

from xicam.gui.widgets.dynimageview import DynImageView
from xicam.gui.widgets.plotwidgetmixins import CurveLabels


class Hint(object):
    canvas_cls = None

    def __init__(self, **kwargs):
        self.parent = None
        self.checked = False
        self.enabled = True
        self._name = None

    def init_canvas(self, **kwargs):
        return self.canvas_cls()

    @property
    def name(self):
        raise NotImplementedError

    def remove(self):
        raise NotImplementedError

    def selective_copy(self, var_mapping: Dict):
        # TODO raise
        return self

    def visualize(self, canvas):
        raise NotImplementedError


class PlotHint(Hint):
    canvas_cls = CurveLabels

    def __init__(self, x: np.ndarray, y: np.ndarray, xLog: bool = False, yLog: bool = False, labels=None, **kwargs):
        super(PlotHint, self).__init__()
        if kwargs.get("name"):
            self._name = kwargs["name"]
        self.x = x
        self.y = y
        self.kwargs = kwargs
        self.item = None
        self.canvas = None
        self.xLog = xLog
        self.yLog = yLog
        self.labels = labels

    def init_canvas(self, addLegend=False, **kwargs):
        canvas = super(PlotHint, self).init_canvas(**kwargs)
        if addLegend:
            canvas.addLegend(offset=(-1, 1))
        return canvas

    @property
    def name(self):
        if not self._name:
            left = self.kwargs.get("labels", dict()).get("left")
            bottom = self.kwargs.get("labels", dict()).get("bottom")
            if left and bottom:
                self._name = f"{left} vs. {bottom}"
            else:
                self._name = "Plot"
        return self._name

    def visualize(self, canvas, color=None):
        plotItem = canvas.plotItem
        plotItem.setLabels(**(self.labels or {}))
        plotItem.setLogMode(x=self.xLog, y=self.yLog)
        if self.kwargs.get("name"):
            self.item = canvas.plot(self.x, self.y, **self.kwargs)
        else:
            self.item = canvas.plot(self.x, self.y, name=self._name, **self.kwargs)
        self.canvas = canvas

        style = self.kwargs.get("style")
        # Update colors according to number of plot data items in the canvas
        if not color:
            numItems = len(plotItem.items)
            for i in range(numItems):
                color = (float(i) / numItems * 255, (1 - float(i) / numItems) * 255, 255)
                plotItem.items[i].setPen(color, **self.kwargs)
        else:
            self.item.setPen(color, **self.kwargs)

    def remove(self):
        legend = self.canvas.plotItem.legend
        if legend:
            # Don't rely on pg.LegendItem.removeItem, as it uses str comparison (instead of refs) to remove!
            # legend.removeItem(self.name)
            for sample, label in legend.items:
                if sample.item is self.item:
                    legend.items.remove((sample, label))
                    legend.layout.removeItem(sample)
                    sample.close()
                    legend.layout.removeItem(label)
                    label.close()
                    legend.updateSize()
        self.canvas.removeItem(self.item)
        self.item = None
        if not list(filter(lambda item: isinstance(item, pg.PlotCurveItem), self.canvas.items())):
            parent_widget = self.canvas.parent().parent()
            parent_widget.removeTab(parent_widget.indexOf(self.canvas))

    def selective_copy(self, var_mapping: Dict):
        new_plot_hint = copy.copy(self)
        new_plot_hint.x = var_mapping[self.x]
        new_plot_hint.y = var_mapping[self.y]
        return new_plot_hint


class VerticalROI(Hint):
    # Design needs to be thought out here, could be either a plot widget or image view
    # TODO -- range should not be a Var
    # canvas_cls = tuple(pg.PlotWidget, pg.ImageView)

    def __init__(self, range: str, **kwargs):
        super(VerticalROI, self).__init__()
        self.range = range
        self.kwargs = kwargs
        self.canvas = None

    @property
    def name(self):
        return f"{self.parent.name} Vertical ROI"

    def visualize(self, canvas):  # TODO: callables?
        from pyqtgraph import LinearRegionItem

        self.canvas = canvas
        self.canvas.addItem(LinearRegionItem(*self.range.value, **self.kwargs))

    @property
    def name(self):
        return f"{self.parent.name} Vertical ROI"


class ImageHint(Hint):
    # TODO -- change to DynImageView (address quickMinMax when resizing the two-time image)
    canvas_cls = DynImageView
    ref_count = count(0)

    def __init__(
        self, image, name="", invertY=False, xlabel: str = None, ylabel: str = None, transform=None, z: int = None, **kwargs
    ):
        self._name = name
        super(ImageHint, self).__init__()
        self.count = next(self.ref_count)
        self.image = image
        self.invertY = invertY
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.transform = transform
        if transform is None:
            transform = QTransform()
            transform.translate(0, -1)
            transform.scale(0, -1)
            self.transform = transform
        self.z = z
        self.kwargs = kwargs
        self.enabled = False
        self.canvas = None

    def init_canvas(self, **kwargs):
        self.canvas = self.canvas_cls(view=pg.PlotItem(labels=dict(left=self.xlabel, bottom=self.ylabel)))
        self.canvas.view.invertY(self.invertY)
        return self.canvas

    @property
    def name(self):
        if not self._name:
            self._name = "Image " + str(self.count)
        return self._name

    def visualize(self, canvas):
        self.canvas = canvas
        self.canvas.setImage(self.image, **self.kwargs)

    def remove(self):
        parent_widget = self.canvas.parent().parent()
        parent_widget.removeTab(parent_widget.indexOf(self.canvas))

    def selective_copy(self, var_mapping: Dict):
        new_plot_hint = copy.copy(self)
        new_plot_hint.image = var_mapping[self.image]
        return new_plot_hint


class CoPlotHint(Hint):
    canvas_cls = CurveLabels
    canvas_map = dict()

    def __init__(self, *plothints: List[PlotHint], **kwargs):
        if kwargs.get("name"):
            self._name = kwargs["name"]
        super(CoPlotHint, self).__init__()
        self.plothints = [*plothints]  # type: List[PlotHint]
        self.kwargs = kwargs
        self.canvas = None

    def init_canvas(self, addLegend=False, **kwargs):
        canvas = super(CoPlotHint, self).init_canvas(**kwargs)
        self.canvas_map[canvas] = 0
        if addLegend:
            canvas.addLegend(offset=(-1, 1))
        return canvas

    @property
    def name(self):
        if not self._name:
            self._name = f"Plot of " + ", ".join([hint.name for hint in self.plothints])
        return self._name

    def remove(self):
        for plothint in self.plothints:
            plothint.remove()
        self.canvas_map[self.canvas] -= 1
        if self.canvas_map[self.canvas] < 0:
            self.canvas_map[self.canvas] = 0

    def visualize(self, canvas):
        self.canvas = canvas
        self.canvas_map[self.canvas] += 1
        numItems = self.canvas_map[self.canvas]
        # Update colors according to number of co-plot hints in the same canvas
        for i in range(numItems):
            color = (float(i) / numItems * 255, (1 - float(i) / numItems) * 255, 255)
        for plothint in self.plothints:
            # TODO: should this rely on the contained plothints' visualize? or should we directly plot on the canvas?
            plothint.visualize(self.canvas, color=color)
            # self.canvas.plot(plothint.x, plothint.y, **{**plothint.kwargs, **self.kwargs})



from .plugin import PluginType

viewTypes = ["ListView", "TreeView", ""]

try:
    from qtpy.QtCore import *

    class DataSourceListModel(QAbstractListModel):
        def __init__(self, dataresource):
            super(DataSourceListModel, self).__init__()
            self.dataresource = dataresource
            self.dataresource.model = self
            self.rowCount = dataresource.rowCount
            self.data = dataresource.data
            self.columnCount = dataresource.columnCount
            self.refresh = dataresource.refresh

        @property
        def config(self):
            return self.dataresource.config

        @property
        def uri(self):
            return self.dataresource.uri

        @uri.setter
        def uri(self, value):
            self.dataresource.uri = value

        def __getattr__(self, attr):  ## implicitly wrap methods from leftViewer
            if hasattr(self.dataresource, attr):
                m = getattr(self.dataresource, attr)
                return m
            raise NameError(attr)


except ImportError:
    # TODO: how should this be handled?
    pass


class DataResourcePlugin(PluginType):
    from xicam.gui.widgets.dataresourcebrowser import DataResourceList, DataBrowser

    model = DataSourceListModel
    view = DataResourceList
    controller = DataBrowser

    is_singleton = False
    needs_qt = True

    name = ""

    def __init__(self, flags: dict = None, **config):
        """
        Config keys should follow RFC 3986 URI format:
            scheme:[//[user[:password]@]host[:port]][/path][?query][#fragment]

        Should provide the abstract methods required of QAbstractItemModel. While this plugin does not depend on Qt, it
        mimics the same functionality, and so can easily be wrapped in a QAbstractItemModel for GUI views. A parent
        model assigns itself to self.model
        """
        super(DataResourcePlugin, self).__init__()
        # self.model = None
        self.config = config
        self.flags = flags if flags else {"isFlat": True, "canPush": False}
        # self.uri=''

        import warnings

        warnings.warn("The DataResourcePlugin is being deprecated in favor of CatalogPlugin.", DeprecationWarning)

    def pushData(self, *args, **kwargs):
        raise NotImplementedError

    def dataChanged(self, topleft=None, bottomright=None):
        if self.model:
            self.model.dataChanged.emit(topleft, bottomright)

    def columnCount(self, index=None):
        raise NotImplementedError

    def rowCount(self, index=None):
        raise NotImplementedError

    def data(self, index, role):
        raise NotImplementedError

    def headerData(self, column, orientation, role):
        raise NotImplementedError

    def index(self, row, column, parent):
        raise NotImplementedError

    def parent(self, index):
        raise NotImplementedError

    @property
    def host(self):
        return self.config["host"]

    @property
    def path(self):
        return self.config["path"]

    def refresh(self):
        pass

    # TODO: convenience properties for each config



from collections import OrderedDict
from enum import Enum
from typing import Dict, List, Union
from xicam.core.data import NonDBHeader
from qtpy.QtGui import *
from databroker.core import BlueskyRun

from .plugin import PluginType


class GUIPlugin(PluginType):
    """GUIPlugin class for interactive Xi-CAM plugins.

    This class represents the fundamental interactive plugin for Xi-CAM.

    GUIPlugins are left uninstanciated until all plugins are loaded so that all dependent widgets are loaded before
    the UI is setup. They DO become singletons.
    """

    is_singleton = True
    needs_qt = True

    def __init__(self):
        """Creates a GUIPlugin.

        When implementing your own GUIPlugin, you will want to set the `self.stages` property
        and then call `super` at the end of `__init__`.
        """
        super(GUIPlugin, self).__init__()
        self.stage = list(self.stages.values())[0]

    def appendHeader(self, header: NonDBHeader, **kwargs):
        # kwargs can include flags for how the data append operation is handled, i.e.:
        #   - as a new doc
        #   - merged into the current doc (stream)
        #   - as a new doc, flattened by some operation (average)
        pass

    def appendCatalog(self, catalog: BlueskyRun, **kwargs):
        """Re-implement to define how to add a catalog to your GUIPlugin.

        Parameters
        ----------
        catalog : BlueskyRun
            Catalog reference that you can use as you wish (corresponds to the opened catalog in DataResourceBrowser).
        """
        # kwargs can include flags for how the data append operation is handled, i.e.:
        #   - as a new doc
        #   - merged into the current doc (stream)
        #   - as a new doc, flattened by some operation (average)
        pass

    def currentheader(self) -> Dict:
        raise NotImplementedError

    @property
    def headers(self) -> List[NonDBHeader]:
        return [self.headermodel.item(i) for i in range(self.headermodel.rowCount())]

    @property
    def stages(self) -> OrderedDict:
        """Returns the stages of the GUIPlugin.

        A stage is defined by a GUILayout; each stage represents a distinct user-interface in a GUIPlugin.
        """
        return self._stages

    @stages.setter
    def stages(self, stages):
        """Set the GUIPlugin's stages"""
        self._stages = stages

    @property
    def exposedvars(self) -> Dict:
        raise NotImplementedError


class PanelState(Enum):
    """
    Represents the state of a panel (widget). Some panels have default
    widgets. The data preview widget and data browser widget are defaults
    for the top-left and left panels respectively. In order to not show
    these widgets, PanelState.Disabled must be set to those panels.

    """

    Disabled = 1
    Defaulted = 2
    Customized = 3


class GUILayout(object):
    """
    Represents a layout of dockable widgets in a 3x3 grid.

    The parameters can either be a PanelState value or a QWidget object. Note
    that only the `center` parameter is required; the other parameters default
    to `PanelState.Defaulted`. The default behavior of a `PanelState.Defaulted`
    widget is to be hidden.

    Parameters
    ----------
    center : Union[QWidget, PanelState]
        The center widget
    left : Union[QWidget, PanelState], optional
        The left widget
    right : Union[QWidget, PanelState], optional
        The right widget
    bottom : Union[QWidget, PanelState], optional
        The bottom widget
    top : Union[QWidget, PanelState], optional
        The top widget
    lefttop : Union[QWidget, PanelState], optional
        The top-left widget
    righttop : Union[QWidget, PanelState], optional
        The top-right widget
    leftbottom : Union[QWidget, PanelState], optional
        The bottom-left widget
    rightbottom : Union[QWidget, PanelState], optional
        The bottom-right widget

    Notes
    -----
    For an example of how this class can be used, see the xicam.gui
    XicamMainWindow class.

    """

    def __init__(
        self,
        center,
        left=PanelState.Defaulted,
        right=PanelState.Defaulted,
        bottom=PanelState.Defaulted,
        top=PanelState.Defaulted,
        lefttop=PanelState.Defaulted,
        righttop=PanelState.Defaulted,
        leftbottom=PanelState.Defaulted,
        rightbottom=PanelState.Defaulted,
    ):
        self.topwidget = top
        self.leftwidget = left
        self.rightwidget = right
        self.centerwidget = center
        self.bottomwidget = bottom
        self.lefttopwidget = lefttop
        self.righttopwidget = righttop
        self.leftbottomwidget = leftbottom
        self.rightbottomwidget = rightbottom

    def __getitem__(self, item: str):
        if not item.endswith("widget"):
            item += "widget"
        return getattr(self, item, PanelState.Defaulted)



from xicam.plugins import OperationPlugin
from typing import Callable, List, Union, Tuple
from collections import defaultdict
from xicam.core import msg, execution
from xicam.core.threads import QThreadFuture, QThreadFutureIterator
from weakref import ref


# TODO: add debug flag that checks mutations by hashing inputs
class Graph(object):
    """Graph that represents operations as nodes and links as edges.

    This class is used as the underlying representation for a Workflow.

    Note that this is similar to a directed graph;
    however, this Graph allows multiple edges between two nodes,
    as there can be multiple outputs for operation A that link to
    multiple inputs for operation B.

    """

    def __init__(self):
        self._operations = []
        self._inbound_links = defaultdict(lambda: defaultdict(lambda: []))
        self._outbound_links = defaultdict(lambda: defaultdict(lambda: []))
        self._disabled_operations = set()

    def add_operation(self, operation: OperationPlugin):
        """Add a single operation into the workflow."""
        self.add_operations(operation)

    def add_operations(self, *operations: OperationPlugin):
        """Add operations into the workflow.

        This will add the list of operations to the end of the workflow.
        """
        # NOTE: Because of some Pycharm bugs, class-decorators break type inspection; expect to get complaints
        #       https://youtrack.jetbrains.com/issue/PY-27142
        for operation in operations:
            self.insert_operation(len(self._operations) + 1, operation)

    def insert_operation(self, index: int, operation: OperationPlugin):
        """Insert an operation at a specific index in the workflow.

        Parameters
        ----------
        index : int
            Index where to insert the operation. 0 will add at the beginning; -1 will add to the end.
        operation : OperationPlugin
            Operation to insert.
        """
        if not isinstance(operation, OperationPlugin):
            raise TypeError(f'Expected "{operation}" to be an OperationPlugin')
        self._operations.insert(index, operation)
        operation._workflow = ref(self)

        self.notify()

    def _validate_link_parameters(self, source, dest, source_param, dest_param):
        if not isinstance(source, OperationPlugin):
            raise TypeError(f'Expected source "{source}" to be an OperationPlugin')
        if not isinstance(dest, OperationPlugin):
            raise TypeError(f'Expected dest "{dest}" to be an OperationPlugin')
        if not isinstance(source_param, str):
            raise TypeError(f'Expected source_param "{source_param}" to be a string')
        if not isinstance(dest_param, str):
            raise TypeError(f'Expected dest_param "{dest_param}" to be a string')
        if source not in self._operations:
            raise ValueError(f'Source operation "{source}" cannot be found')
        if dest not in self._operations:
            raise ValueError(f'Destination operation "{dest}" cannot be found')

    def add_link(self, source, dest, source_param, dest_param):
        """Add a link between two operations in the workflow.

        Links are defined from an operation's parameter to another operation's parameter.
        This creates a connection between two operations during execution of a workflow.

        Parameters
        ----------
        source : OperationPlugin
            The operation to link from.
        dest : OperationPlugin
            The operation to link to.
        source_param : str
            Name of the parameter in the source operation to link (source of the data; output).
        dest_param : str
            Name of the parameter in the destination operation to link (where the data goes; input).
        """
        self._validate_link_parameters(source, dest, source_param, dest_param)

        if source_param not in source.output_names:
            raise ValueError(f'An output named "{source_param}" could not be found in the source operation, {source.name}')
        elif dest_param not in dest.input_names:
            raise ValueError(f'An input named "{dest_param}" could not be found in the destination operation, {dest.name}')
        self._inbound_links[dest][source].append((source_param, dest_param))
        self._outbound_links[source][dest].append((source_param, dest_param))

        self.notify()

    def remove_link(self, source, dest, source_param, dest_param):
        """Remove a link between two operations.

        Parameters
        ----------
        source : OperationPlugin
            The source operation to remove a link from.
        dest : OperationPlugin
            The destination operation to remove a link from.
        source_param : str
            Name of the source parameter that is defining the link to remove.
        dest_param : str
            Name of the destination parameter that is defining the link to remove.
        """
        self._validate_link_parameters(source, dest, source_param, dest_param)

        inbound_links = self._inbound_links[dest][source]
        outbound_links = self._outbound_links[source][dest]
        link = (source_param, dest_param)
        if link not in inbound_links:
            raise ValueError(f"There is no inbound link '{link}' to remove for source '{source}', dest '{dest}'")
        if link not in outbound_links:
            raise ValueError(f"There is no outbound link '{link}' to remove for source '{source}', dest '{dest}'")
        inbound_links.remove((source_param, dest_param))
        if len(inbound_links) == 0:
            del self._inbound_links[dest]
        outbound_links.remove((source_param, dest_param))
        if len(outbound_links) == 0:
            del self._outbound_links[source]

        self.notify()

    def get_inbound_links(self, operation):
        """Returns the links connected to the operation given (linked inputs of the operation).

        The returned dict represents all operations that are connected to `operation`.
        Links are represented as a list of 2-element tuples,
        where the first element of the tuple is another operation's output parameter,
        and the second element of the tuple is `operation`'s input parameter.

        Using `keys()` will give all of the operations that connect to `operation`.
        Using `values()` will give all of the links from each operation to `operation`.

        Parameters
        ----------
        operation : OperationPlugin
            Operation to get incoming links for (some operation -> `operation`).

        Returns
        -------
        defaultdict
            Returns a dictionary defining all of the links from any connected operations to `operation`.
        """
        if not isinstance(operation, OperationPlugin):
            raise TypeError(f"Expected an operation (got {operation} of type {type(operation)}")
        try:
            return self._inbound_links[operation]
        except KeyError as e:
            raise KeyError(f"Operation {operation} (named {operation.name} not found in inbound links")

    def get_outbound_links(self, operation):
        """Returns the links connected from the operation given (linked outputs of the operation).

        The returned dict represents all the operations that `operation` connects to.
        Links are represented as a list of 2-element tuples,
        where the first element of the tuple is `operation`'s output parameter,
        and the second element of the tuple is another operation's input parameter.

        Using `.keys()` on the returned dict will give all of the operations that `operation` connects to.
        Using `.values()` on the returned dict will give all of the links from `operation` to each operation.

        Parameters
        ----------
        operation : OperationPlugin
            Operation to get outgoing links for (`operation` -> some operation).

        Returns
        -------
        defaultdict
            Returns a dictionary defining all of the links from `operation` to any connected operations.
        """
        if not isinstance(operation, OperationPlugin):
            raise TypeError(f"Expected an operation (got {operation} of type {type(operation)}")
        try:
            return self._outbound_links[operation]
        except KeyError as e:
            raise KeyError(f"Operation {operation} (named {operation.name} not found in outbound links")

    def clear_operation_links(self, operation, clear_outbound=True, clear_inbound=True):
        """Remove all links for an operation."""

        # TODO : clear_outbound, clear_inbound
        if not isinstance(operation, OperationPlugin):
            raise TypeError(f"Expected '{operation}' to be an OperationPlugin")
        try:
            self._inbound_links[operation].clear()
            self._outbound_links[operation].clear()

            for op, value in self._inbound_links.items():
                if operation in self._inbound_links[op]:
                    del self._inbound_links[op][operation]

            for op, value in self._outbound_links.items():
                if operation in self._outbound_links[op]:
                    del self._outbound_links[op][operation]

            self.notify()

        except KeyError as e:
            raise KeyError(f"Operation {operation} (named {operation}) not found")

    def clear_links(self):
        """Remove all links from the workflow, but preserve the operations."""
        self._inbound_links.clear()
        self._outbound_links.clear()

        self.notify()

    def clear_operations(self):
        """Remove all operations and links from the workflow."""
        self._operations.clear()
        self._inbound_links.clear()
        self._outbound_links.clear()

        self.notify()

    def remove_operation(self, operation, remove_orphan_links=True):
        """Remove an operation from the workflow.

        Parameters
        ----------
        operation : OperationPlugin
            Operation to remove from the workflow.
        remove_orphan_links : bool
            If True, removes all links that link to the operation to be removed.
            If False, does not remove any links for the operation
            and returns the removed operations links dict (default is True).

        Returns
        -------
        defaultdict
            By default (`remove_orphan_links` is True), returns None.
            Otherwise, returns the links for the removed operation.
        """
        if not isinstance(operation, OperationPlugin):
            raise TypeError(f"Expected an operation (got {operation} of type {type(operation)}")

        try:
            self._operations.remove(operation)
        except ValueError as e:
            raise ValueError(f"Operation {operation} (named {operation.name} found")

        if remove_orphan_links:
            self.clear_operation_links(operation)
        else:
            return self.get_inbound_links(operation), self.get_outbound_links(operation)

        self.notify()

    def _end_operations(self):
        """
        find tasks at the end of the graph and work up
        check inputs and remove dependency nodes, what is left is unique ones
        """

        end_tasks = set(self.operations) - self._outbound_links.keys()

        msg.logMessage("End tasks:", *[task.name for task in end_tasks], msg.DEBUG)
        return end_tasks

    def _dask_graph(self):
        dask_graph = {}

        for operation in self.operations:
            links = {}
            dependent_ids = []
            for dep_operation, inbound_links in self._inbound_links[operation].items():
                for (source_param, dest_param) in inbound_links:
                    links[dest_param] = source_param
                dependent_ids.append(dep_operation.id)

            node = _OperationWrapper(operation, links)
            dask_graph[operation.id] = (node, *dependent_ids)

        return dask_graph

    def as_dask_graph(self):
        """
        process from end tasks and into all dependent ones

        Returns a tuple that represents the graph as a dask-compatible
        graph for processing. The second element of the tuple identifies
        the end node ids (i.e. nodes that do not have connected outputs).

        Returns
        -------
        tuple
            A tuple with two-elements, the first being the dask graph,
            the second being the end task ids.
        """

        for (i, node) in enumerate(self.operations):
            node.id = str(i)

        end_tasks = self._end_operations()

        dask_graph = self._dask_graph()
        end_task_ids = [i.id for i in end_tasks]

        return dask_graph, end_task_ids

    @property
    def operations(self):
        """Returns the operations of this workflow."""
        return self._operations

    def links(self):
        """Returns all the links defined in the workflow.

        Returns a list of tuples, each tuple representing a link as follows:
        source operation, destination operation, source parameter, destination parameter.

        Note that the links are shown as outbound links.

        Returns
        -------
        list
            Returns a list of the links (defined as outbound links) in the workflow.

        """
        return [
            (src, dest, link[0], link[1])
            for src, outbound_links in self._outbound_links.items()
            for dest, links in outbound_links.items()
            for link in links
        ]

    def operation_links(self, operation: OperationPlugin):
        """Returns the outbound links for an operation.

        Returns a list of tuples, each tuple representing a link as follows:
        `operation`, destination operation, `operation` source parameter, destination parameter.

        Returns
        -------
        list
            Returns a list of the links (defined as outbound links) for `operation`.
        """
        if not isinstance(operation, OperationPlugin):
            raise TypeError(f"Expected {operation} to be an OperationPlugin")
        if operation not in self.operations:
            raise ValueError(f'Operation "{operation}" not found')
        return [
            (operation, dest, link[0], link[1]) for dest, links in self._outbound_links[operation].items() for link in links
        ]

    def disabled_operations(self):
        """Returns the disabled operations (if any) in the workflow."""
        return self._disabled_operations

    def disabled(self, operation):
        """Indicate if the operation is disabled in the workflow.

        Parameters
        ----------
        operation : OperationPlugin
            Operation to check if it is disabled or not.

        Returns
        -------
        bool
            Returns True if the operation is disabled in the Workflow; otherwise False.
        """
        return operation in self._disabled_operations

    def set_disabled(
        self, operation: OperationPlugin, value: bool = True, remove_orphan_links: bool = True, auto_connect_all: bool = True
    ):
        """Set an operation's disabled state in the workflow.

        By default when disabling an operation, links connected to the operation will be removed
        (`remove_orphan_links` would be True).
        If `value` is False (re-enabling an operation), then no links are changed.

        Parameters
        ----------
        operation : OperationPlugin
            The operation whose disabled state is being modified.
        value : bool
            Indicates the disabled state (default is True, which disables the operation).
        remove_orphan_links : bool
            If True and `value` is True, removes the links connected to the operation.
            Otherwise, no links are changed (default is True).
        auto_connect_all : bool
            If True, then a best-effort attempt will be made to try to reconnect the
            operations in the workflow (default is True).
            See the `Graph.auto_connect_all` method for more information.
        Returns
        -------
        list
            Returns a list of any orphaned links for an operation that is set to disabled.
            Default behavior will return an empty list (when `remove_orphan_links` is True).
            If enabling an operation (`value` is False), then an empty list is returned,
            as no links are changed.
        """
        if value:
            self._disabled_operations.add(operation)
        else:
            if operation in self._disabled_operations:
                self._disabled_operations.remove(operation)

        orphaned_links = []
        if value:
            if remove_orphan_links:
                self.clear_operation_links(operation)
            else:
                orphaned_links = self.operation_links(operation)

        if auto_connect_all:
            self.auto_connect_all()

        self.notify()
        return orphaned_links

    def toggle_disabled(self, operation: OperationPlugin, remove_orphan_links=True, auto_connect_all=True):
        """Toggle the disable state of an operation.

        By default, when an operation is toggled to a disabled state,
        any links connected to the operation will be removed.

        Parameters
        ----------
        operation : OperationPlugin
            The operation to toggle disable state for.
        remove_orphan_links : bool
            If True, when the operation's toggle state is toggled to disabled,
            any links connected to the operation will be removed (default is True).

        Returns
        -------
        list
            Returns a list of any orphaned links for an operation.
            Default behavior will return an empty list.
            A non-empty list can be returned when `remove_orphan_links` is False
            and the connected operation is toggled to disabled.

        """
        is_disabled = False
        if operation in self._disabled_operations:
            is_disabled = True
        return self.set_disabled(operation, not is_disabled, remove_orphan_links, auto_connect_all)

    def auto_connect_all(self):
        """Attempts to automatically connect operations together by matching output names and input names.

        Makes a best-effort to link operations based on the names of their outputs and inputs.
        If operation A has an output named "image", and operation B has an input named "image",
        then A "image" will link to B "image".
        Outputs and inputs that have matching types in addition to matching names
        will be favored more for the auto-connection.

        If there are no outputs with matching inputs (by name), no links will be added.
        """
        self.clear_links()

        # for each operation
        for input_operation in self.operations:

            # for each input of given operation
            for input_name in input_operation.input_names:
                bestmatch = None
                matchness = 0
                # Parse backwards from the given operation, looking for matching outputs
                for output_operation in reversed(self.operations[: self.operations.index(input_operation)]):
                    # check each output
                    for output_name in output_operation.output_names:
                        # if matching name
                        if output_name == input_name:
                            # if a name match hasn't been found
                            if matchness < 1:
                                bestmatch = output_operation, output_name
                                matchness = 1
                                # if a name+type match hasn't been found
                                if output_operation.output_types[output_name] == input_operation.input_types[input_name]:
                                    if matchness < 2:
                                        bestmatch = output_operation, output_name
                                        matchness = 2
                if bestmatch:
                    self.add_link(bestmatch[0], input_operation, bestmatch[1], input_name)
                    msg.logMessage(
                        f"connected {bestmatch[0].name}:{bestmatch[1]} to {input_operation.name}:{input_name}",
                        level=msg.DEBUG,
                    )

    def notify(self, *args, **kwargs):
        """See Workflow.notify"""
        pass

    def _pretty_print(self):
        """Print out links in easy-to-read format."""

        def print_underlined(text):
            underline = "-" * len(text)
            print(f"{text}\n{underline}")

        print_underlined("Graph:")
        # Assumes a link has the following structure:
        # tuple(source operation, destination operation, source parameter, destination parameter)
        for link in self.links():
            source = link[0]
            dest = link[1]
            source_param = link[2]
            dest_param = link[3]
            print(f"\t({source.name})[{source_param}] ---> [{dest_param}]({dest.name})")

        print_underlined("Inbound links:")
        for op, link in self._inbound_links.items():
            for source, l in dict(link).items():
                print(f"\t({op.name}) from ({source.name}) via {l}")

        print_underlined("Outbound links:")
        for op, link in self._outbound_links.items():
            for dest, l in dict(link).items():
                print(f"\t({op.name}) to ({dest.name}) via {l}")


class _OperationWrapper:
    def __init__(self, node, named_args, islocal=False):
        self.node = node
        self.named_args = named_args
        self.islocal = islocal
        self.queues_in = {}
        self.queues_out = {}

        self.node.__internal_data__ = self

    # args = [{'name':value}]

    def __call__(self, *args):
        node_args = {}
        for arg, (input_name, sender_operation_name) in zip(args, self.named_args.items()):
            node_args[input_name] = arg[sender_operation_name]

        result_keys = self.node.output_names
        result_values = self.node(**node_args)
        if not isinstance(result_values, tuple):
            result_values = (result_values,)

        return dict(zip(result_keys, result_values))

    def __repr__(self):
        # return getattr(self.node, "name", self.node.__class__.__name__)
        return self.node.__class__.__name__


class Workflow(Graph):
    def __init__(self, name="", operations=None):
        """
        Create a Workflow that can be executed.

        Parameters
        ----------
        name : str, optional
            Name of the Workflow (default is "").
        operations : List, optional
            List of operations to add to the Workflow being created (default is None).
        """
        super(Workflow, self).__init__()
        # self._operations = []  # type: List[OperationPlugin]
        self._observers = set()
        # self._links = []  # type: List[Tuple[ref, str, ref, str]]
        if name:
            self.name = name

        if operations:
            # self._operations.extend(operations)
            self.add_operations(*operations)
        self.staged = False

        self.lastresult = []

    def stage(self, connection):
        """
        Stages required data resources to the compute resource. Connection will be a Connection object (WIP) keeping a
        connection to a compute resource, include connection.hostname, connection.username...

        Returns
        -------
        QThreadFuture
            A concurrent.futures-like qthread to monitor status. Returns True if successful
        """
        self.staged = True
        # TODO: Operations invalidate parent workflow staged attribute if data resources are modified, but not parameters
        # TODO: check if data is accessible from compute resource; if not -> copy data to compute resource
        # TODO: use cam-link to mirror installation of plugin packages

    def execute(
        self,
        executor=None,
        connection=None,
        callback_slot=None,
        finished_slot=None,
        except_slot=None,
        default_exhandle=True,
        lock=None,
        fill_kwargs=True,
        threadkey=None,
        **kwargs,
    ):
        """
        Execute this workflow on the specified host. Connection will be a Connection object (WIP) keeping a connection
        to a compute resource, include connection.hostname, connection.username...

        Returns
        -------
        QThreadFuture
            A concurrent.futures-like qthread to monitor status. The future's callback_slot receives the result.

        """
        if not self.staged:
            self.stage(connection)

        if fill_kwargs:
            self.fill_kwargs(**kwargs)

        if executor is None:
            executor = execution.executor

        future = QThreadFuture(
            executor.execute,
            self,
            callback_slot=callback_slot,
            finished_slot=finished_slot,
            except_slot=except_slot,
            default_exhandle=default_exhandle,
            lock=lock,
            threadkey=threadkey,
        )
        future.start()
        return future

    def execute_synchronous(self, executor=None, connection=None, fill_kwargs=True, **kwargs):
        if not self.staged:
            self.stage(connection)

        if fill_kwargs:
            self.fill_kwargs(**kwargs)

        if executor is None:
            executor = execution.executor

        return executor.execute(self)

    def execute_all(
        self,
        connection=None,
        executor=None,
        callback_slot=None,
        finished_slot=None,
        except_slot=None,
        default_exhandle=True,
        lock=None,
        fill_kwargs=True,
        threadkey=None,
        **kwargs,
    ):
        """
        Execute this workflow on the specified host. Connection will be a Connection object (WIP) keeping a connection
        to a compute resource, include connection.hostname, connection.username...

        Each kwargs is expected to be an iterable of the same length; these values will be iterated over, zipped, and
        executed through the workflow.

        Returns
        -------
        QThreadFuture
            A concurrent.futures-like qthread to monitor status. The future's callback_slot receives the result.

        """
        if not self.staged:
            self.stage(connection)

        if executor is None:
            executor = execution.executor

        def executeiterator(workflow):
            for kwargvalues in zip(*kwargs.values()):
                zipkwargs = dict(zip(kwargs.keys(), kwargvalues))
                if fill_kwargs:
                    self.fill_kwargs(**zipkwargs)
                yield (executor.execute)(workflow)

        future = QThreadFutureIterator(
            executeiterator,
            self,
            callback_slot=callback_slot,
            finished_slot=finished_slot,
            except_slot=except_slot,
            default_exhandle=default_exhandle,
            lock=lock,
            threadkey=threadkey,
        )
        future.start()
        return future

    def fill_kwargs(self, **kwargs):
        """
        Fills in all empty inputs with names matching keys in kwargs.
        """
        for operation in self.operations:
            for key in kwargs:
                if key in operation.input_names:
                    operation.filled_values[key] = kwargs[key]

    def validate(self):
        """
        Validate all of:\
        - All required inputs are satisfied.
        - Connection is active.
        - ?

        Returns
        -------
        bool
            True if workflow is valid.

        """
        # TODO: add validation
        return True

    def attach(self, observer: Callable):
        """Add an observer to the Workflow.

        An observer is a callable that is called when the Workflow.notify
        method is called. In other words,
        the observer will be called whenever the Workflow state changes;
        for example, links are modified, operations are removed, etc.
        When notified, the observer is called.

        Parameters
        ----------
        observer : Callable
            A callable to add from the Workflow.
        """
        self._observers.add(observer)

    def detach(self, observer: Callable):
        """Remove an observer from the Workflow.

        An observer is a callable that is called when the Workflow.notify
        method is called. In other words,
        the observer will be called whenever the Workflow state changes;
        for example, links are modified, operations are removed, etc.
        When notified, the observer is called.

        Parameters
        ----------
        observer : Callable
            The callable to remove from the Workflow.
        """
        if observer in self._observers:
            self._observers.remove(observer)

    def notify(self):
        """Notify the observers; the observers will be called.
        """
        for observer in self._observers:
            observer()

    @property
    def hints(self):
        hints = []
        for operation in self._operations:
            hints.extend(operation.hints)
        return hints

    def visualize(self, canvas, **canvases):
        canvasinstances = {name: canvas() if callable(canvas) else canvas for name, canvas in canvases.items()}
        for operation in self._operations:
            for hint in operation.hints:
                hint.visualize(canvas)



import sys, os
from xicam.core.data import load_header
from urllib import parse

if "qtpy" in sys.modules:
    from qtpy.QtWidgets import *
    from qtpy.QtCore import QSettings, QDir

    class LocalFileSystemResourcePlugin(QFileSystemModel):
        def __init__(self):
            super(LocalFileSystemResourcePlugin, self).__init__()

            self.uri = parse.urlparse(QSettings().value("lastlocaldir", os.getcwd()))
            self.setResolveSymlinks(True)

            self.setRootPath(parse.urlunparse(self.uri))

        def setRootPath(self, path):
            if os.path.isdir(path):
                filter = "*"
                root = path
            else:
                filter = os.path.basename(path)
                root = path[: -len(filter)]

            root = QDir(root)
            super(LocalFileSystemResourcePlugin, self).setRootPath(root.absolutePath())
            self.setNameFilters([filter])
            self.uri = parse.urlparse(path)
            QSettings().setValue("lastlocaldir", path)

        @property
        def path(self):
            return self.rootPath()

        @path.setter
        def path(self, value):
            self.setRootPath(value)

        def refresh(self):
            self.setRootPath(parse.urlunparse(self.uri))

        def getHeader(self, indexes):
            uris = [self.filePath(index) for index in indexes]
            return load_header(uris=uris)



import sys
import os
import platform
import itertools
import warnings

import entrypoints
from appdirs import user_config_dir, site_config_dir, user_cache_dir

from xicam.core import msg
from xicam.core import threads
from xicam.core.args import parse_args

from .datahandlerplugin import DataHandlerPlugin
from .catalogplugin import CatalogPlugin
from .guiplugin import GUIPlugin, GUILayout
from .operationplugin import OperationPlugin
from .settingsplugin import SettingsPlugin, ParameterSettingsPlugin
from .dataresourceplugin import DataResourcePlugin
from .controllerplugin import ControllerPlugin
from .widgetplugin import QWidgetPlugin
from .plugin import PluginType

try:
    # try to find the venvs entrypoint
    if "cammart" in entrypoints.get_group_named(f"xicam.plugins.SettingsPlugin") and not "--no-cammart" in sys.argv:
        from xicam.gui.cammart.venvs import observers as venvsobservers
        from xicam.gui.cammart import venvs
    else:
        raise ImportError
except ImportError:
    venvsobservers = None
from .dataresourceplugin import DataResourcePlugin
from .fittablemodelplugin import Fittable1DModelPlugin
from .ezplugin import _EZPlugin, EZPlugin
from .hints import PlotHint, Hint

from queue import LifoQueue
from enum import Enum, auto
from contextlib import contextmanager
from timeit import default_timer

import time

op_sys = platform.system()
if op_sys == "Darwin":  # User config dir incompatible with venv on darwin (space in path name conflicts)
    user_plugin_dir = os.path.join(user_cache_dir(appname="xicam"), "plugins")
else:
    user_plugin_dir = os.path.join(user_config_dir(appname="xicam"), "plugins")
site_plugin_dir = os.path.join(site_config_dir(appname="xicam"), "plugins")



@contextmanager
def load_timer():
    start = default_timer()
    elapser = lambda: default_timer() - start
    yield lambda: elapser()
    end = default_timer()
    elapser = lambda: end - start


class State(Enum):
    READY = auto()
    DISCOVERING = auto()
    LOADING = auto()
    INSTANTIATING = auto()


class Filters(Enum):
    UPDATE = auto()
    COMPLETE = auto()


class XicamPluginManager:
    def __init__(self, qt_is_safe=False):

        self.qt_is_safe = qt_is_safe
        self._blacklist = []
        self._load_queue = LifoQueue()
        self._instantiate_queue = LifoQueue()
        self._entrypoints = {}
        self._load_cache = {}
        self._observers = []
        self.state = State.READY
        self.type_mapping = {}
        self.plugin_types = {}

        # Remember all modules loaded before any plugins are loaded; don't bother unloading these
        self._preloaded_modules = set(sys.modules.keys())

        # Observe changes to venvs
        if venvsobservers is not None:
            venvsobservers.append(self)

        self.initialize_types()

        # Check if cammart should be ignored
        try:
            args = parse_args(exit_on_fail=False)
            include_cammart = not args.nocammart
            self._blacklist = args.blacklist
        except RuntimeError:
            include_cammart = False

        # ...if so, blacklist it
        if not include_cammart:
            self._blacklist.extend(["cammart", "venvs"])

    def initialize_types(self):
        # Load plugin types
        self.plugin_types = {name: ep.load() for name, ep in entrypoints.get_group_named("xicam.plugins.PluginType").items()}

        # Toss plugin types that need qt if running without qt
        if not self.qt_is_safe:
            self.plugin_types = {
                type_name: type_class
                for type_name, type_class in self.plugin_types.items()
                if not getattr(type_class, "needs_qt", True)
            }

        # Initialize types
        self.type_mapping = {type_name: {} for type_name in self.plugin_types.keys()}
        self._entrypoints = {type_name: {} for type_name in self.plugin_types.keys()}
        self._load_cache = {type_name: {} for type_name in self.plugin_types.keys()}

    def collect_plugins(self):
        """
        Find, load, and instantiate all Xi-cam plugins matching known plugin types

        """
        self._discover_plugins()
        self._load_plugins()

    def collect_plugin(self, plugin_name, plugin_class, type_name, replace=False):
        """
        Register a class as a plugin. For in-memory usage. If `replace`, then any earlier instances are purged first

        """
        if replace:
            # Clear cache by name
            self._entrypoints[type_name].pop(plugin_name, None)
            self.type_mapping[type_name].pop(plugin_name, None)
            self._load_cache[type_name].pop(plugin_name, None)
        else:
            try:
                assert plugin_name not in self.type_mapping[type_name]
            except AssertionError:
                raise ValueError(f"A plugin named {plugin_name} has already been loaded. Supply `replace=True` to override.")

        # Start a special collection cycle
        self.state = State.DISCOVERING
        live_entry_point = LiveEntryPoint(plugin_name, plugin_class)
        self._load_queue.put((type_name, live_entry_point))
        if self.state == State.DISCOVERING:
            self.state = State.LOADING
        self._load_plugins()

    def _unload_plugins(self):
        assert self.state == State.READY
        self._load_queue = LifoQueue()
        self._instantiate_queue = LifoQueue()

        # Initialize types
        self.type_mapping = {type_name: {} for type_name in self.plugin_types.keys()}
        self._entrypoints = {type_name: {} for type_name in self.plugin_types.keys()}
        self._load_cache = {type_name: {} for type_name in self.plugin_types.keys()}

        reload_candidates = list(filter(lambda key: key.startswith("xicam."), sys.modules.keys()))
        for module_name in reload_candidates:
            if module_name not in self._preloaded_modules:
                del sys.modules[module_name]

    def hot_reload(self):
        warnings.warn("Hot-reloading plugins; unexpected and unpredictable behavior may occur...", UserWarning)
        self._unload_plugins()
        self.collect_plugins()

    def _discover_plugins(self):
        self.state = State.DISCOVERING
        # for each plugin type
        for type_name in self.plugin_types.keys():

            # get all entrypoints matching that group
            group = entrypoints.get_group_named(f"xicam.plugins.{type_name}")
            group_all = entrypoints.get_group_all(f"xicam.plugins.{type_name}")

            # check for duplicate names
            self._check_shadows(group, group_all)

            for name, entrypoint in group.items():
                # If this entrypoint hasn't already been queued
                if entrypoint not in self._entrypoints[type_name] and entrypoint.name not in self._blacklist:
                    # ... queue and cache it
                    self._load_queue.put((type_name, entrypoint))
                    self._entrypoints[type_name][name] = entrypoint

            msg.logMessage(f"Discovered {type_name} entrypoints:", *self._entrypoints[type_name].values(), sep="\n")
        if self.state == State.DISCOVERING:
            self.state = State.LOADING

    @staticmethod
    def _check_shadows(group, group_all):
        # Warn the user if entrypoint names may shadow each other
        if len(group_all) != len(group):
            # There are some name collisions. Let's go digging for them.
            for name, matches in itertools.groupby(group_all, lambda ep: ep.name):
                matches = list(matches)
                if len(matches) != 1:
                    winner = group[name]
                    warnings.warn(
                        f"There are {len(matches)} conflicting entrypoints which share the name {name!r}:\n{matches}\n"
                        f"Loading entrypoint from {winner.module_name} and ignoring others."
                    )

    @threads.method(
        threadkey="entrypoint-loader", showBusy=False, cancelIfRunning=False
    )  # progress state managed independently
    def _load_plugins(self):
        started_instantiating = False

        # For every entrypoint in the load queue
        while not self._load_queue.empty():
            type_name, entrypoint = self._load_queue.get()

            # load it
            self._load_plugin(type_name, entrypoint)

            if not started_instantiating:  # If this is the first load
                # Start an event chain to pull from the queue
                threads.invoke_as_event(self._instantiate_plugin)
                started_instantiating = True

            # mark it as completed
            self._load_queue.task_done()

        # Finished loading, progress
        if self.state == State.LOADING:
            self.state = State.INSTANTIATING

    def _load_plugin(self, type_name, entrypoint: entrypoints.EntryPoint):
        # if the entrypoint was already loaded into cache and queued, do nothing
        if self._load_cache[type_name].get(entrypoint.name, None):
            return

        try:
            # Load the entrypoint (unless already cached), cache it, and put it on the instantiate queue
            msg.logMessage(f"Loading entrypoint {entrypoint.name} from module: {entrypoint.module_name}")
            with load_timer() as elapsed:
                plugin_class = self._load_cache[type_name][entrypoint.name] = (
                    self._load_cache[type_name].get(entrypoint.name, None) or entrypoint.load()
                )
        except (Exception, SystemError) as ex:
            msg.logMessage(f"Unable to load {entrypoint.name} plugin from module: {entrypoint.module_name}", msg.ERROR)
            msg.logError(ex)
            msg.notifyMessage(
                repr(ex), title=f'An error occurred while starting the "{entrypoint.name}" plugin.', level=msg.CRITICAL
            )

        else:
            msg.logMessage(f"{int(elapsed() * 1000)} ms elapsed while loading {entrypoint.name}", level=msg.INFO)
            self._instantiate_queue.put((type_name, entrypoint, plugin_class))

    def _instantiate_plugin(self):
        if not self._instantiate_queue.empty():
            type_name, entrypoint, plugin_class = self._instantiate_queue.get()

            # if this plugin was already instantiated earlier, skip it; mark done
            if self.type_mapping[type_name].get(entrypoint.name, None) is None:

                # inject the entrypoint name into the class
                plugin_class._name = entrypoint.name

                success = False

                # ... and instantiate it (as long as its supposed to be singleton)

                try:
                    if getattr(plugin_class, "is_singleton", False):
                        msg.logMessage(f"Instantiating {entrypoint.name} plugin object.", level=msg.INFO)
                        with load_timer() as elapsed:
                            self.type_mapping[type_name][entrypoint.name] = plugin_class()

                        msg.logMessage(
                            f"{int(elapsed() * 1000)} ms elapsed while instantiating {entrypoint.name}", level=msg.INFO
                        )
                    else:
                        self.type_mapping[type_name][entrypoint.name] = plugin_class
                    success = True

                except (Exception, SystemError) as ex:
                    msg.logMessage(
                        f"Unable to instantiate {entrypoint.name} plugin from module: {entrypoint.module_name}", msg.ERROR
                    )
                    msg.logError(ex)
                    msg.notifyMessage(repr(ex), title=f'An error occurred while starting the "{entrypoint.name}" plugin.')

                if success:
                    msg.logMessage(f"Successfully collected {entrypoint.name} plugin.", level=msg.INFO)
                    msg.showProgress(self._progress_count(), maxval=self._entrypoint_count())
                    self._notify(Filters.UPDATE)

            # mark it as completed
            self._instantiate_queue.task_done()

        # If this was the last plugin
        if self._load_queue.empty() and self._instantiate_queue.empty() and self.state in [State.INSTANTIATING, State.READY]:
            self.state = State.READY
            msg.logMessage("Plugin collection completed!")
            msg.hideProgress()
            self._notify(Filters.COMPLETE)

        if not self.state == State.READY:  # if we haven't reached the last task, but there's nothing queued
            threads.invoke_as_event(self._instantiate_plugin)  # return to the event loop, but come back soon

    def _get_plugin_by_name(self, name, type_name):
        return_plugin = None
        # Check all types matching type_name
        for search_type_name in self.plugin_types.keys():
            if type_name == search_type_name or not type_name:
                match_plugin = self.type_mapping[search_type_name].get(name, None)
                if match_plugin and return_plugin:
                    raise ValueError(
                        "Multiple plugins with the same name but different types exist. " "Must specify type_name."
                    )
                return_plugin = match_plugin

        return return_plugin

    def _get_entrypoint_by_name(self, name, type_name):
        return_entrypoint = None
        return_type = None
        # Check all types matching type_name
        for search_type_name in self.plugin_types.keys():
            if type_name == search_type_name or not type_name:
                match_entrypoint = self._entrypoints[search_type_name].get(name, None)
                if match_entrypoint and return_entrypoint:
                    raise ValueError(
                        "Multiple plugins with the same name but different types exist. " "Must specify type_name."
                    )
                return_entrypoint = match_entrypoint
                return_type = search_type_name

        return return_entrypoint, return_type

    def get_plugin_by_name(self, name, type_name=None, timeout=10):
        """
        Find a collected plugin named `name`, optionally by also specifying the type of plugin.

        Parameters
        ----------
        name : str
            name of the plugin to get
        type_name : str
            type of the plugin to get (optional)

        Returns
        -------
        object
            the matching plugin object (may be a class or instance), or None if not found
        """
        return_plugin = self._get_plugin_by_name(name, type_name)

        if return_plugin:
            return return_plugin

        # If still actively collecting plugins
        if self.state != State.READY:
            # find the matching entrypoint
            entrypoint, type_name = self._get_entrypoint_by_name(name, type_name)

            if not entrypoint:
                raise NameError(
                    f"The plugin named {name} of type {type_name} could not be discovered. "
                    f"Check your installation integrity."
                )

            # Load it immediately; it will move to top of instantiate queue as well
            msg.logMessage(f"Immediately loading {entrypoint.name}.", level=msg.INFO)
            self._load_plugin(type_name, entrypoint)

            # Add another instantiate event to the Qt event queue, so that it triggers in the next event loop
            threads.invoke_as_event(self._instantiate_plugin)

            # wait for it to load
            with load_timer() as elapsed:
                while not return_plugin:
                    return_plugin = self._get_plugin_by_name(name, type_name)
                    if threads.is_main_thread():
                        from qtpy.QtWidgets import QApplication  # Required as late import to avoid loading Qt things too soon
                        QApplication.processEvents()
                    else:
                        time.sleep(0.01)
                    if elapsed() > timeout:
                        raise TimeoutError(f"Plugin named {name} waited too long to instantiate and timed out")

        return return_plugin

    def get_plugins_of_type(self, type_name):
        return list(self.type_mapping[type_name].values())

    def attach(self, callback, filter=None):
        """
        Subscribe a callback to receive notifications. If a filter is used, only matching notifications are sent.
        See `Filters` for options.

        """
        self._observers.append((callback, filter))

    def _notify(self, filter=None):
        """ Notify all observers. Observers attached with filters much mach the emitted filter to be notified."""
        for callback, obsfilter in self._observers:
            if obsfilter == filter or not obsfilter:
                callback()

    def venvChanged(self):
        self.collect_plugins()

    def _entrypoint_count(self):
        return sum(map(len, self._entrypoints.values()))

    def _progress_count(self):
        return sum(map(len, self.type_mapping.values()))

    def getPluginsOfCategory(self, type_name):
        raise NotImplementedError("This method has been renamed to follow snake_case")
        warnings.warn("Transition to snake_case in progress...", DeprecationWarning)
        return self.get_plugins_of_type(type_name)

    def collectPlugins(self):
        raise NotImplementedError("This method has been renamed to follow snake_case")
        warnings.warn("Transition to snake_case in progress...", DeprecationWarning)
        return self.collect_plugins()

    def getPluginByName(self, plugin_name, type_name):
        raise NotImplementedError("This method has been renamed to follow snake_case")
        warnings.warn("Transition to snake_case in progress...", DeprecationWarning)
        return self.get_plugin_by_name(plugin_name, type_name)


# Setup plugin manager
manager = XicamPluginManager()


# A light class to mimic EntryPoint for live objects
class LiveEntryPoint(entrypoints.EntryPoint):
    def __init__(self, name, object, extras=None, distro=None):
        super(LiveEntryPoint, self).__init__(
            name, module_name="[live]", object_name=object.__name__, extras=extras, distro=distro
        )
        self.object = object

    def load(self):
        return self.object


from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



from appdirs import user_config_dir, site_config_dir, user_cache_dir
import os
import platform


user_cache_dir = user_cache_dir(appname="xicam")
site_config_dir = site_config_dir(appname="xicam")
user_config_dir = user_config_dir(appname="xicam")
user_dev_dir = os.path.expanduser("~/Xi-cam/plugins")
op_sys = platform.system()
if op_sys == "Darwin":  # User config dir incompatible with venv on darwin (space in path name conflicts)
    user_plugin_dir = os.path.join(user_cache_dir, "plugins")
else:
    user_plugin_dir = os.path.join(user_config_dir, "plugins")
site_plugin_dir = os.path.join(site_config_dir, "plugins")


def init_dir(path):
    try:
        os.makedirs(path)
    except FileExistsError:
        pass


def init_dirs():
    for path in [user_cache_dir, user_config_dir, user_plugin_dir, user_dev_dir]:
        init_dir(path)



from qtpy.QtWidgets import QScrollArea, QLabel

from xicam.gui.static import path


class MOTD(QScrollArea):
    def __init__(self, parent=None):
        super(MOTD, self).__init__()
        text = f"""
        <div align='center'>
           <br />
           <br />
           <img src='{path('images/camera.jpg')}' width='200'/>
           <h1 style='font-family:Zero Threes;'>
               Welcome to Xi-cam
           </h1>
           <br />
           Please cite Xi-cam in published work: <br />
           <a href="http://scripts.iucr.org/cgi-bin/paper?S1600577518005787" style="color:white;">Pandolfi, R. J., et al. (2018). J. Synchrotron Rad. <b>25</b>, 1261-1270.</a>
    
        </div>
        
        <table cellspacing="10"><tr><td>
            Ronald J. Pandolfi<sup>a</sup><br />
            Daniel B. Allan<sup>e</sup> <br />
            Elke Arenholz<sup>a</sup> <br />
            Luis Barroso-Luque<sup>a</sup><br />
            Stuart I. Campbell<sup>e</sup><br />
            Thomas A. Caswell<sup>e</sup><br />
            Austin Blair<sup>a</sup><br />
            Francesco De Carlo<sup>c</sup><br />
            Sean Fackler<sup>a</sup><br />
            Amanda P. Fournier<sup>b</sup><br />
            Guillaume Freychet<sup>a</sup><br />
            Masafumi Fukuto<sup>e</sup><br />
            Dŏga G̈ursoy<sup>ch</sup><br />
            Zhang Jiang<sup>c</sup><br />
            Harinarayan Krishnan<sup>a</sup><br />
            Dinesh Kumar<sup>a</sup><br />
            R. Joseph Kline<sup>g</sup><br />
            Ruipeng Li<sup>e</sup><br />
            Christopher Liman<sup>g</sup><br />
        </td><td>
            Stefano Marchesini<sup>a</sup><br />
            Apurva Mehta<sup>b</sup><br />
            Alpha T. N’Diaye<sup>a</sup><br />
            Dilworth (Dula) Y. Parkinson<sup>a</sup><br />
            Holden Parks<sup>a</sup><br />
            Lenson A. Pellouchoud<sup>a</sup><br />
            Talita Perciano<sup>a</sup><br />
            Fang Ren<sup>b</sup><br />
            Shreya Sahoo<sup>a</sup><br />
            Joseph Strzalka<sup>c</sup><br />
            Daniel Sunday<sup>g</sup><br />
            Christopher J. Tassone<sup>a</sup><br />
            Daniela Ushizima<sup>a</sup><br />
            Singanallur Venkatakrishnan<sup>d</sup><br />
            Kevin G. Yager<sup>f</sup><br />
            James A. Sethian<sup>a</sup><br />
            Alexander Hexemer<sup>a</sup>
        </td><td>
            <sup>a</sup>Lawrence Berkeley National Laboratory<br />
            <sup>b</sup>Stanford Synchrotron Radiation Lightsource<br />
            <sup>c</sup>Advanced Photon Source, Argonne National Laboratory<br />
            <sup>d</sup>Oak Ridge National Laboratory<br />
            <sup>e</sup>National Synchrotron Light Source II<br />
            <sup>f</sup>Center for Functional Nanomaterials<br />
            <sup>g</sup>National Institute of Standards and Technology<br />
            <sup>h</sup>Department of Electrical Engineering<br />and Computer Science, Northwestern University<br />
        </td></tr></table>
        
        """

        label = QLabel()
        label.setText(text)
        self.setWidget(label)



from pytestqt import qtbot
from xicam.core.tests.workflow_fixtures import simple_workflow, square_op, sum_op


def test_daskflow(qtbot, simple_workflow):


    from pyqtgraph.flowchart import Node, Terminal

    # COREDUMPS HERE!
    a = Terminal(Node("test1"), "test1", "in")

    from xicam.gui.widgets.daskflow import DaskFlow

    df = DaskFlow()
    df.fromDask(simple_workflow)

    qtbot.addWidget(df)


from qtpy.QtWidgets import (
    QTabWidget,
    QHBoxLayout,
    QVBoxLayout,
    QToolBar,
    QMenu,
    QAction,
    QTreeView,
    QListView,
    QWidget,
    QSizePolicy,
    QTabBar,
)
from qtpy.QtCore import QObject, QAbstractItemModel, QSize, Qt, QEvent, Signal, QSettings
from qtpy.QtGui import QIcon, QPixmap, QKeyEvent
from intake.catalog.base import Catalog
from intake.catalog.entry import CatalogEntry
from ..clientonlymodels.LocalFileSystemResource import LocalFileSystemResourcePlugin
from xicam.gui.static import path
from xicam.core.data import NonDBHeader, load_header


from xicam.core import threads
from .searchlineedit import SearchLineEdit
from urllib import parse
from pathlib import Path
from functools import partial
import os, webbrowser
from xicam.gui.widgets.tabview import ContextMenuTabBar
from xicam.gui.bluesky.databroker_catalog_plugin import DatabrokerCatalogPlugin


class BrowserTabWidget(QTabWidget):
    def __init__(self, parent=None):
        super(BrowserTabWidget, self).__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)


class DataBrowser(QWidget):
    sigOpen = Signal(NonDBHeader)
    sigPreview = Signal(NonDBHeader)

    def __init__(self, browserview):
        super(DataBrowser, self).__init__()

        hbox = QHBoxLayout()
        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(0)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(0)
        self.setContentsMargins(0, 0, 0, 0)

        self.browserview = browserview
        self.browserview.sigOpen.connect(self.sigOpen)
        self.browserview.sigPreview.connect(self.sigPreview)
        self.browserview.sigOpenExternally.connect(self.openExternally)
        self.browserview.sigURIChanged.connect(self.uri_to_text)
        self.toolbar = QToolBar()
        self.toolbar.addAction(QIcon(QPixmap(str(path("icons/up.png")))), "Move up directory", self.moveUp)
        # self.toolbar.addAction(QIcon(QPixmap(str(path('icons/filter.png')))), 'Filter')
        self.toolbar.addAction(QIcon(QPixmap(str(path("icons/refresh.png")))), "Refresh", self.hardRefreshURI)
        self.toolbar.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.URILineEdit = SearchLineEdit("", clearable=False)
        self.uri_to_text()

        hbox.addWidget(self.toolbar)
        hbox.addWidget(self.URILineEdit)
        vbox.addLayout(hbox)
        vbox.addWidget(self.browserview)
        self.setLayout(vbox)

        self.URILineEdit.textChanged.connect(self.softRefreshURI)
        self.URILineEdit.returnPressed.connect(self.softRefreshURI)  # hard refresh
        self.URILineEdit.focusOutEvent = self.softRefreshURI  # hard refresh

        self.hardRefreshURI()

    def text_to_uri(self):
        uri = parse.urlparse(self.URILineEdit.text())
        self.browserview.model().uri = uri
        return uri

    def uri_to_text(self):
        uri = self.browserview.model().uri
        text = parse.urlunparse(uri)
        self.URILineEdit.setText(text)
        return text

    def hardRefreshURI(self, *_, **__):
        self.text_to_uri()
        self.browserview.refresh()

    def moveUp(self):
        self.browserview.model().uri = parse.urlparse(str(Path(self.URILineEdit.text()).parent))
        self.browserview.refresh()
        self.uri_to_text()

    def openExternally(self, uri):
        webbrowser.open(uri)

    softRefreshURI = hardRefreshURI


class BrowserTabBar(ContextMenuTabBar):
    sigAddBrowser = Signal(object, str)

    def __init__(self, tabwidget: QTabWidget):
        super(BrowserTabBar, self).__init__()

        self.tabwidget = tabwidget
        self.tabwidget.setTabBar(self)

        self.setExpanding(False)
        self.setTabsClosable(True)

        plusPixmap = QPixmap(str(path("icons/plus.png")))
        self.plusIcon = QIcon(plusPixmap)
        # self.plus_button.setToolTip('Open a new browser')
        # self.plus_button.setParent(self)
        # self.plus_button.setMaximumSize(32, 32)
        # self.plus_button.setMinimumSize(32, 32)
        # self.plus_button.clicked.connect(self.sigAddBrowser.emit)
        tab = self.addTab(self.plusIcon, "")
        try:
            self.tabButton(tab, QTabBar.RightSide).resize(0, 0)
            self.tabButton(tab, QTabBar.RightSide).hide()
        except AttributeError:
            self.tabButton(tab, QTabBar.LeftSide).resize(0, 0)
            self.tabButton(tab, QTabBar.LeftSide).hide()
        # self.movePlusButton()  # Move to the correct location
        # self.setDocumentMode(True)
        self.currentChanged.connect(self.tabwidget.setCurrentIndex)
        self.currentChanged.connect(self.saveLastTab)
        self.installEventFilter(self)

    def saveLastTab(self, i):
        if i < 2:
            QSettings().setValue("databrowsertab", i)

    def addTab(self, *args, **kwargs):
        return self.insertTab(self.count() - 1, *args, **kwargs)

    def eventFilter(self, object, event):
        try:
            if (
                object == self
                and event.type() in [QEvent.MouseButtonPress, QEvent.MouseButtonRelease]
                and event.button() == Qt.LeftButton
            ):

                if event.type() == QEvent.MouseButtonPress:
                    tabIndex = object.tabAt(event.pos())
                    if tabIndex == self.count() - 1:
                        self.showMenu(self.mapToGlobal(event.pos()))
                        return True
            return False
        except Exception as e:
            print("Exception raised in eventfilter", e)

    def showMenu(self, pos):
        self.menu = QMenu()
        # Add all resource plugins
        self.actions = {}
        from xicam.plugins import manager as pluginmanager

        for plugin in pluginmanager.get_plugins_of_type("CatalogPlugin") + pluginmanager.get_plugins_of_type(
            "DataResourcePlugin"
        ):
            self.actions[plugin.name] = QAction(plugin.name)
            self.actions[plugin.name].triggered.connect(partial(self._addBrowser, plugin))
            self.menu.addAction(self.actions[plugin.name])

        self.menu.popup(pos)

    def _addBrowser(self, plugin):
        from xicam.plugins import DataResourcePlugin, CatalogPlugin

        plugin = plugin()

        if isinstance(plugin, DataResourcePlugin):
            datasource = plugin
            self.sigAddBrowser.emit(datasource.controller(datasource.view(datasource.model(datasource))), datasource.name)
        elif isinstance(plugin, Catalog):
            self.sigAddBrowser.emit(plugin.controller, plugin.name)


class DataResourceView(QObject):
    def __init__(self, model: QAbstractItemModel):
        super(DataResourceView, self).__init__()
        self._model = model  # type: QAbstractItemModel
        self.setModel(self._model)
        self.doubleClicked.connect(self.open)
        self.setSelectionMode(self.ExtendedSelection)
        self.setSelectionBehavior(self.SelectRows)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.menuRequested)

        self.menu = QMenu()
        standardActions = [
            QAction("Open", self),
            QAction("Open Externally", self),
            QAction("Enable/Disable Streaming", self),
            QAction("Delete", self),
        ]
        self.menu.addActions(standardActions)
        standardActions[0].triggered.connect(self.open)
        standardActions[1].triggered.connect(self.openExternally)

    def menuRequested(self, position):
        self.menu.exec_(self.viewport().mapToGlobal(position))

    def open(self, index):
        pass

    def currentChanged(self, current, previous):
        pass

    def openExternally(self, uri: str):
        pass


class DataResourceTree(QTreeView, DataResourceView):
    sigOpen = Signal(object)
    sigOpenPath = Signal(str)
    sigOpenExternally = Signal(str)
    sigPreview = Signal(object)
    sigURIChanged = Signal()

    def __init__(self, *args):
        super(DataResourceTree, self).__init__(*args)

    def refresh(self):
        self.model().refresh()
        self.setRootIndex(self.model().index(self.model().path))


class DataResourceList(QListView, DataResourceView):
    sigOpen = Signal(object)
    sigOpenPath = Signal(str)
    sigOpenExternally = Signal(str)
    sigPreview = Signal(object)
    sigURIChanged = Signal()

    def refresh(self):
        self.model().refresh()

    @threads.method()
    def open(self, _):
        indexes = self.selectionModel().selectedRows()
        if len(indexes) == 1:
            path = os.path.join(self.model().config["path"], self.model().data(indexes[0], Qt.DisplayRole).value())
            if self.model().isdir(indexes[0]):
                self.model().config["path"] = path
                self.sigURIChanged.emit()
                return
        uris = [self.model().pull(index) for index in indexes]
        header = load_header(uris=uris)
        if header:
            self.sigOpen.emit(header)


class LocalFileSystemTree(DataResourceTree):
    def __init__(self):
        super(LocalFileSystemTree, self).__init__(LocalFileSystemResourcePlugin())

    def open(self, _=None):
        indexes = self.selectionModel().selectedRows()
        if len(indexes) == 1:
            path = self.model().filePath(indexes[0])
            if os.path.isdir(path):
                self.model().path = path
                self.setRootIndex(indexes[0])
                self.sigURIChanged.emit()
                return
        self.sigOpen.emit(self.model().getHeader(indexes))

    def currentChanged(self, current, previous):
        if current.isValid():
            header = self.model().getHeader([current])
            if header:
                self.sigPreview.emit(header)

        self.scrollTo(current)

    def keyPressEvent(self, event: QKeyEvent):
        super(LocalFileSystemTree, self).keyPressEvent(event)
        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            event.accept()
            self.open()

    def openExternally(self, uri: str):
        indexes = self.selectionModel().selectedRows()
        for index in indexes:
            self.sigOpenExternally.emit(self.model().filePath(index))


class DataResourceBrowser(QWidget):
    sigOpen = Signal(NonDBHeader)
    sigPreview = Signal(NonDBHeader)

    def __init__(self):
        super(DataResourceBrowser, self).__init__()
        vbox = QVBoxLayout()
        vbox.setSpacing(0)
        vbox.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(QSize(250, 400))

        # cache pre-load index (otherwise it gets overwritten during init
        index = int(QSettings().value("databrowsertab", 0))

        self.browsertabwidget = BrowserTabWidget(self)
        self.browsertabbar = BrowserTabBar(self.browsertabwidget)
        self.browsertabbar.sigAddBrowser.connect(self.addBrowser)
        self.browsertabbar.tabCloseRequested.connect(self.closetab)

        # Add the required 'Local' browser
        self.addBrowser(DataBrowser(LocalFileSystemTree()), "Local", closable=False)
        self.addBrowser(DatabrokerCatalogPlugin().controller, "Databroker", closable=False)
        self.browsertabbar.setCurrentIndex(index)

        vbox.addWidget(self.browsertabwidget)

        self.setLayout(vbox)

    def closetab(self, i):
        if hasattr(self.browsertabwidget.widget(i), "closable"):
            if self.browsertabwidget.widget(i).closable:
                self.browsertabwidget.removeTab(i)

    def sizeHint(self):
        return QSize(250, 400)

    def addBrowser(self, databrowser: DataBrowser, text: str, closable: bool = True):
        databrowser.sigOpen.connect(self.sigOpen)
        databrowser.sigPreview.connect(self.sigPreview)
        databrowser.closable = closable
        tab = self.browsertabwidget.addTab(databrowser, text)
        self.browsertabwidget.setCurrentIndex(tab)
        # self.browsertabbar.addTab(text)
        if closable is False:
            try:
                self.browsertabbar.tabButton(tab, QTabBar.RightSide).resize(0, 0)
                self.browsertabbar.tabButton(tab, QTabBar.RightSide).hide()
            except AttributeError:
                self.browsertabbar.tabButton(tab, QTabBar.LeftSide).resize(0, 0)
                self.browsertabbar.tabButton(tab, QTabBar.LeftSide).hide()



from collections import OrderedDict

import qdarkstyle
from qtmodern import styles
from qtpy.QtGui import *
from qtpy.QtWidgets import QApplication
from xicam.gui.static import path
import pyqtgraph as pg

from xicam.plugins import ParameterSettingsPlugin


def setDefault():
    QApplication.instance().setStyleSheet("")


def setDark():
    QApplication.instance().setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())


def setModern():
    styles.dark(QApplication.instance())


def setUglyGreen():
    QApplication.instance().setStyleSheet("QWidget {background-color: darkgreen;}")


def setPlotWhite():
    pg.setConfigOption("background", "w")
    pg.setConfigOption("foreground", "k")


def setPlotDefault():
    pass


class AppearanceSettingsPlugin(ParameterSettingsPlugin):
    def __init__(self):
        super(AppearanceSettingsPlugin, self).__init__(
            QIcon(str(path("icons/colors.png"))),
            "Appearance",
            [
                dict(
                    name="Theme",
                    values=OrderedDict(
                        [
                            ("Default", setDefault),
                            ("QDarkStyle", setDark),
                            ("QtModern", setModern),
                            ("UglyGreen", setUglyGreen),
                        ]
                    ),
                    type="list",
                ),
                dict(
                    name="Plot Theme (requires restart)",
                    values=OrderedDict([("Default (Dark)", setPlotDefault), ("Publication (White)", setPlotWhite)]),
                    type="list",
                ),
            ],
        )

    def apply(self):
        self["Theme"]()
        self["Plot Theme (requires restart)"]()



from qtpy.QtWidgets import QLabel
from xicam.plugins import GUIPlugin, GUILayout
from xicam.gui.widgets.dynimageview import DynImageView
from xicam.gui.widgets.imageviewmixins import XArrayView
from xicam.core.data import MetaXArray


class TestPlugin(GUIPlugin):
    name = "catalogtest"

    def __init__(self):
        self.imageview = XArrayView()

        self.stages = {
            "Stage 1": GUILayout(self.imageview),
        }

        super(TestPlugin, self).__init__()

    def appendCatalog(self, runcatalog, **kwargs):
        xdata = runcatalog().primary.read()["random_img"][:, :, :, 0]  # The test data is 4-dimensional; ignoring last dim
        self.imageview.setImage(MetaXArray(xdata))

    def appendHeader(self):
        ...



from collections import OrderedDict

import qdarkstyle
from qtmodern import styles
from qtpy.QtGui import *
from qtpy.QtWidgets import QApplication
from xicam.gui.static import path
import pyqtgraph as pg
from xicam import plugins
from xicam.core.execution import localexecutor, daskexecutor, camlinkexecutor
from xicam.core import execution

from xicam.plugins import ParameterSettingsPlugin


class ExecutionSettingsPlugin(ParameterSettingsPlugin):
    def __init__(self):
        super(ExecutionSettingsPlugin, self).__init__(
            QIcon(str(path("icons/cpu.png"))),
            "Execution",
            [
                dict(
                    name="Executor",
                    values=OrderedDict(
                        [
                            ("Local Threaded", localexecutor.LocalExecutor()),
                            ("Local Service", daskexecutor.DaskExecutor()),
                            ("Cam-link", None),
                        ]
                    ),
                    value="Local Threaded",
                    type="list",
                )
            ],
        )

        execution.executor = self["Executor"]

    def apply(self):
        execution.executor = self["Executor"]



import sys
from qtpy.QtCore import QTimer, Qt
from qtpy.QtGui import QMovie, QPixmap
from qtpy.QtWidgets import QSplashScreen, QApplication

from xicam.gui import static


def elide(s: str, max_len: int = 60):
    if len(s) > max_len:
        s = s[: max_len - 3] + "..."
    return s


class XicamSplashScreen(QSplashScreen):
    minsplashtime = 5000

    def __init__(self, log_path: str, initial_length: int, f: int = Qt.WindowStaysOnTopHint | Qt.SplashScreen):
        """
        A QSplashScreen customized to display an animated gif. The splash triggers launch when clicked.

        After minsplashtime, this splash waits until the animation finishes before triggering the launch.

        Parameters
        ----------
        log_path    :   str
            Path to the Xi-CAM log file to reflect
        initial_length: int
            Length in bytes to seek forward before reading
        f           :   int
            Extra flags (see base class)
        """

        # Get logo movie from relative path
        self.movie = QMovie(str(static.path("images/animated_logo.gif")))

        # Setup drawing
        self.movie.frameChanged.connect(self.paintFrame)
        self.movie.jumpToFrame(1)
        self.pixmap = QPixmap(self.movie.frameRect().size())
        super(XicamSplashScreen, self).__init__(self.pixmap, f)
        self.setMask(self.pixmap.mask())
        self.movie.finished.connect(self.restartmovie)
        self.showMessage("Starting Xi-CAM...")

        self._launching = False
        self._launchready = False
        self.timer = QTimer(self)

        self.log_file = open(log_path, "r")
        self.log_file.seek(initial_length)

        # Start splashing
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.show()
        self.raise_()
        self.activateWindow()
        QApplication.instance().setActiveWindow(self)

        # Setup timed triggers for launching the QMainWindow
        self.timer.singleShot(self.minsplashtime, self.launchwindow)

    def showMessage(self, message: str, color=Qt.darkGray):
        # attempt to parse out everyting besides the message
        try:
            message=message.split(" - ")[-1]
        except Exception:
            pass
        else:
            super(XicamSplashScreen, self).showMessage(elide(message), color=color, alignment=Qt.AlignBottom)

    def mousePressEvent(self, *args, **kwargs):
        # TODO: Apparently this doesn't work?
        self.timer.stop()
        self.execlaunch()

    def show(self):
        """
        Start the animation when shown
        """
        super(XicamSplashScreen, self).show()
        self.movie.start()

    def paintFrame(self):
        """
        Paint the current frame
        """
        self.pixmap = self.movie.currentPixmap()
        self.setMask(self.pixmap.mask())
        self.setPixmap(self.pixmap)
        self.movie.setSpeed(self.movie.speed() + 20)

        line = self.log_file.readline().strip()
        if line:
            self.showMessage(elide(line.split(">")[-1]))

    def sizeHint(self):
        return self.movie.scaledSize()

    def restartmovie(self):
        """
        Once the animation reaches the end, check if its time to launch, otherwise restart animation
        """
        if self._launchready:
            self.execlaunch()
            return
        self.movie.start()

    def launchwindow(self):
        """
        Save state, defer launch until animation finishes
        """
        self._launchready = True

    def execlaunch(self):
        """
        Launch the mainwindow
        """
        if not self._launching:
            self._launching = True

            self.timer.stop()

            # Stop splashing
            self.hide()
            self.movie.stop()
            self.close()
            QApplication.instance().quit()


if __name__ == "__main__":
    qapp = QApplication([])
    splash = XicamSplashScreen(sys.argv[-2], int(sys.argv[-1]))
    qapp.exec_()



from qtpy.QtGui import QIcon

from xicam.core import msg
from xicam.plugins.settingsplugin import ParameterSettingsPlugin
from xicam.gui.static import path


class LoggingSettingsPlugin(ParameterSettingsPlugin):
    """Settings plugin for logging information and parameterization.
    """

    def __init__(self):
        def msg_levels(recommended=""):
            """Returns a dictionary mapping logging level names to their respective integer values.

            Note that `msg.levels` gives us the reverse mapping that we want, from level to level name.

            Parameters
            ----------
            recommended
                Optional string which will mark a logging level as recommended (for use in the list parameter)
                (default is "", which will not mark any levels).

            Returns
            -------
                Dictionary that maps log level names to their values, optionally with one name marked as recommended.

            """
            levels = dict()  # {v: k for k, v in msg.levels.items()}
            for level, level_name in msg.levels.items():
                if recommended and recommended == level_name:
                    levels[level_name + " (recommended)"] = level
                else:
                    levels[level_name] = level
            return levels

        super(LoggingSettingsPlugin, self).__init__(
            QIcon(str(path("icons/log.png"))),
            msg.LOGGING_SETTINGS_NAME,
            [
                # Show users where the log directory is, don't let them modify it though
                dict(
                    name="Log Directory",
                    value=msg.log_dir,
                    type="str",
                    readonly=True,
                    tip="Location where Xi-CAM writes its logs to.",
                ),
                # Allow users to configure the default log level for the xicam logger's FileHandler
                dict(
                    name=msg.FILE_LOG_LEVEL_SETTINGS_NAME,
                    values=msg_levels(recommended="DEBUG"),
                    value=msg.DEFAULT_FILE_LOG_LEVEL,
                    type="list",
                    tip="Changes how much information is logged to the log file in 'Log Directory.'",
                ),
                # Allow users to configure the default log level for the xicam logger's StreamHandler
                dict(
                    name=msg.STREAM_LOG_LEVEL_SETTINGS_NAME,
                    values=msg_levels(),
                    value=msg.DEFAULT_STREAM_LOG_LEVEL,
                    type="list",
                    tip="Changes how much information is logged to the system console / terminal.",
                ),
            ],
        )
        msg.file_handler.setLevel(self[msg.FILE_LOG_LEVEL_SETTINGS_NAME])
        msg.stream_handler.setLevel(self[msg.STREAM_LOG_LEVEL_SETTINGS_NAME])

    def apply(self):
        msg.file_handler.setLevel(self[msg.FILE_LOG_LEVEL_SETTINGS_NAME])
        msg.stream_handler.setLevel(self[msg.STREAM_LOG_LEVEL_SETTINGS_NAME])



from qtpy.QtCore import Qt
from qtpy.QtGui import QFontMetrics, QPainter
from qtpy.QtWidgets import QApplication, QLabel


class ElidedLabel(QLabel):
    def setText(self, text):
        self._text = text
        metrics = QFontMetrics(self.font())
        elided = metrics.elidedText(self._text, Qt.ElideRight, self.width())
        super(ElidedLabel, self).setText(elided)
